import logging

from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings
from django.utils import translation

from pfx.pfxcore.test import APIClient, TestAssertMixin

logger = logging.getLogger(__name__)

test_settings = dict(
    LANGUAGE_CODE='en-us',
    LANGUAGES=[
        ('fr', "French"),
        ('fr-ch', "Swiss French"),
        ('en-us', "English"),
    ],
    USE_I18N=True,
    USE_L10N=True)


class LocaleAPITest(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient()

    @override_settings(**test_settings)
    def test_get_langs(self):
        response = self.client.get('/api/locales/languages')

        self.assertRC(response, 200)
        self.assertJE(response, "meta.count", len(settings.LANGUAGES))

    @override_settings(**test_settings)
    def test_default_lang(self):
        """Test default language is taken."""
        response = self.client.get('/api/test-i18n')

        self.assertEqual(response.headers['Content-Language'], 'en-US')
        self.assertEqual(translation.get_language(), 'en-us')
        self.assertJE(response, "Monday", 'Monday')

    @override_settings(**test_settings)
    def test_browser_accept_language(self):
        """Test accept-language is taken."""
        response = self.client.get(
            '/api/test-i18n',
            HTTP_ACCEPT_LANGUAGE='fr-CH,fr;q=0.9,en-US;q=0.8,en;q=0.7')

        self.assertEqual(response.headers['Content-Language'], 'fr-CH')
        self.assertEqual(translation.get_language(), 'fr-ch')
        self.assertJE(response, "Monday", 'lundi')

    @override_settings(**test_settings)
    def test_browser_undef_accept_language(self):
        """Test fist accept-language is ignored if it is not available."""
        response = self.client.get(
            '/api/test-i18n',
            HTTP_ACCEPT_LANGUAGE=(
                'it-CH,it;q=0.9,fr-CH,fr;q=0.8,en-US;q=0.7,en;q=0.6'))

        self.assertEqual(response.headers['Content-Language'], 'fr-CH')
        self.assertEqual(translation.get_language(), 'fr-ch')
        self.assertJE(response, "Monday", 'lundi')

    @override_settings(**test_settings)
    def test_pfx_custom_language(self):
        """Test x-custom-language has precedence over accept-language."""
        response = self.client.get(
            '/api/test-i18n',
            HTTP_X_CUSTOM_LANGUAGE='fr_CH',
            HTTP_ACCEPT_LANGUAGE='en-US;q=0.9,en;q=0.8')

        self.assertEqual(response.headers['Content-Language'], 'fr-CH')
        self.assertEqual(translation.get_language(), 'fr-ch')
        self.assertJE(response, "Monday", 'lundi')



from django.contrib.auth.models import Group
from django.test import TestCase
from django.test.utils import override_settings

from pfx.pfxcore.apps import update_groups_permissions
from pfx.pfxcore.test import TestAssertMixin


class TestPostMigrateGroupsUpdate(TestAssertMixin, TestCase):

    def test_groups(self):
        update_groups_permissions(dict(
            reader=[
                'tests.view_author'],
            editor=[
                'tests.view_author',
                'tests.add_author']), 'tests', set(), True, 'tests', False)

        groups = {g.name: g for g in Group.objects.all()}
        self.assertEqual(groups.keys(), {'reader', 'editor'})
        perms = {p.codename for p in groups['reader'].permissions.all()}
        self.assertEqual(perms, {'view_author'})
        perms = {p.codename for p in groups['editor'].permissions.all()}
        self.assertEqual(perms, {'view_author', 'add_author'})

        update_groups_permissions(dict(
            editor=[
                'tests.view_author',
                'tests.change_author'],
            new=[]), 'tests', set(), True, 'tests', False)

        groups = {g.name: g for g in Group.objects.all()}
        self.assertEqual(groups.keys(), {'editor', 'new'})
        perms = {p.codename for p in groups['editor'].permissions.all()}
        self.assertEqual(perms, {'view_author', 'change_author'})
        perms = {p.codename for p in groups['new'].permissions.all()}
        self.assertEqual(perms, set())

    @override_settings(PFX_AUTH_GROUPS_CREATE_ONLY=True)
    def test_groups_create_only(self):
        update_groups_permissions(dict(
            reader=[
                'tests.view_author'],
            editor=[
                'tests.view_author',
                'tests.add_author']), 'tests', set(), True, 'tests', False)

        groups = {g.name: g for g in Group.objects.all()}
        self.assertEqual(groups.keys(), {'reader', 'editor'})
        perms = {p.codename for p in groups['reader'].permissions.all()}
        self.assertEqual(perms, {'view_author'})
        perms = {p.codename for p in groups['editor'].permissions.all()}
        self.assertEqual(perms, {'view_author', 'add_author'})

        update_groups_permissions(dict(
            editor=[
                'tests.view_author',
                'tests.change_author'],
            new=[]), 'tests', set(), True, 'tests', False)

        groups = {g.name: g for g in Group.objects.all()}
        self.assertEqual(groups.keys(), {'reader', 'editor', 'new'})
        perms = {p.codename for p in groups['reader'].permissions.all()}
        self.assertEqual(perms, {'view_author'})
        perms = {p.codename for p in groups['editor'].permissions.all()}
        self.assertEqual(perms, {'view_author', 'add_author'})
        perms = {p.codename for p in groups['new'].permissions.all()}
        self.assertEqual(perms, set())



from datetime import date, timedelta
from unittest.mock import MagicMock, patch

from django.core.exceptions import ValidationError
from django.db import connection
from django.test import TestCase

from pfx.pfxcore.fields import MinutesDurationField
from pfx.pfxcore.test import APIClient, TestAssertMixin
from tests.models import Author, Book


class TestFieldsMinutesDuration(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')
        with connection.cursor() as cursor:
            cursor.execute("create extension if not exists unaccent;")

    @classmethod
    def setUpTestData(cls):
        cls.author = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        cls.book = Book.objects.create(
            author=cls.author,
            name="The Hobbit",
            pub_date=date(1937, 1, 1)
        )

    @patch('pfx.pfxcore.fields.logger', MagicMock())
    def test_to_python(self):
        d = MinutesDurationField()
        self.assertIsNone(d.to_python(None))
        self.assertIsNone(d.to_python(''))
        self.assertEqual(d.to_python(0), timedelta(0))
        self.assertEqual(d.to_python('0'), timedelta(0))
        self.assertEqual(
            d.to_python(timedelta(hours=2, minutes=30)),
            timedelta(hours=2, minutes=30))
        self.assertEqual(d.to_python(3), timedelta(hours=3))
        self.assertEqual(d.to_python(2.5), timedelta(hours=2, minutes=30))
        self.assertEqual(d.to_python(0.5), timedelta(minutes=30))
        self.assertEqual(d.to_python("2h 30m"), timedelta(hours=2, minutes=30))
        self.assertEqual(d.to_python("0h 30m"), timedelta(minutes=30))
        self.assertEqual(d.to_python("4:30"), timedelta(hours=4, minutes=30))
        self.assertEqual(d.to_python("0:03"), timedelta(minutes=3))
        self.assertEqual(d.to_python(":03"), timedelta(minutes=3))
        self.assertEqual(d.to_python(":30"), timedelta(minutes=30))
        self.assertEqual(d.to_python("4.5h"), timedelta(hours=4, minutes=30))
        self.assertEqual(d.to_python("4.5"), timedelta(hours=4, minutes=30))
        self.assertEqual(d.to_python("4"), timedelta(hours=4))
        self.assertEqual(d.to_python("4.0"), timedelta(hours=4))
        with self.assertRaises(ValidationError):
            d.to_python(dict(wrong='value'))  # test invalid object type
            d.to_python('a wrong value')

    def test_to_json(self):
        d = MinutesDurationField()
        self.assertIsNone(d.to_json(None))
        self.assertEqual(d.to_json(timedelta(0)), dict(
            minutes=0,
            clock_format='0:00',
            human_format=''))
        self.assertEqual(
            d.to_json(timedelta(hours=2, minutes=30)), dict(
                minutes=150,
                clock_format='2:30',
                human_format='2h\xa030m'))

    def test_minutes_duration_field(self):
        self.book.read_time = timedelta(hours=2, minutes=30)
        self.book.save()
        response = self.client.get(f'/api/books/{self.book.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 150)
        self.assertJE(response, 'read_time.clock_format', "2:30")
        self.assertJE(response, 'read_time.human_format', "2h\xa030m")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time=""))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time', None)

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time=None))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time', None)

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time=0))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 0)
        self.assertJE(response, 'read_time.clock_format', "0:00")
        self.assertJE(response, 'read_time.human_format', "")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="0"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 0)
        self.assertJE(response, 'read_time.clock_format', "0:00")
        self.assertJE(response, 'read_time.human_format', "")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="0:00"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 0)
        self.assertJE(response, 'read_time.clock_format', "0:00")
        self.assertJE(response, 'read_time.human_format', "")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time=2.5))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 150)
        self.assertJE(response, 'read_time.clock_format', "2:30")
        self.assertJE(response, 'read_time.human_format', "2h\xa030m")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="2.5"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 150)
        self.assertJE(response, 'read_time.clock_format', "2:30")
        self.assertJE(response, 'read_time.human_format', "2h\xa030m")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="2.5h"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 150)
        self.assertJE(response, 'read_time.clock_format', "2:30")
        self.assertJE(response, 'read_time.human_format', "2h\xa030m")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="2h 30m"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 150)
        self.assertJE(response, 'read_time.clock_format', "2:30")
        self.assertJE(response, 'read_time.human_format', "2h\xa030m")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="2:30"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 150)
        self.assertJE(response, 'read_time.clock_format', "2:30")
        self.assertJE(response, 'read_time.human_format', "2h\xa030m")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time=2))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 120)
        self.assertJE(response, 'read_time.clock_format', "2:00")
        self.assertJE(response, 'read_time.human_format', "2h")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="2"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 120)
        self.assertJE(response, 'read_time.clock_format', "2:00")
        self.assertJE(response, 'read_time.human_format', "2h")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="2h"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 120)
        self.assertJE(response, 'read_time.clock_format', "2:00")
        self.assertJE(response, 'read_time.human_format', "2h")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="2:00"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 120)
        self.assertJE(response, 'read_time.clock_format', "2:00")
        self.assertJE(response, 'read_time.human_format', "2h")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time=0.5))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 30)
        self.assertJE(response, 'read_time.clock_format', "0:30")
        self.assertJE(response, 'read_time.human_format', "30m")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="0.5"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 30)
        self.assertJE(response, 'read_time.clock_format', "0:30")
        self.assertJE(response, 'read_time.human_format', "30m")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="0.5h"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 30)
        self.assertJE(response, 'read_time.clock_format', "0:30")
        self.assertJE(response, 'read_time.human_format', "30m")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="30m"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 30)
        self.assertJE(response, 'read_time.clock_format', "0:30")
        self.assertJE(response, 'read_time.human_format', "30m")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time=":30"))
        self.assertRC(response, 200)
        self.assertJE(response, 'read_time.minutes', 30)
        self.assertJE(response, 'read_time.clock_format', "0:30")
        self.assertJE(response, 'read_time.human_format', "30m")

        response = self.client.put(
            f'/api/books/{self.book.pk}', dict(read_time="INVALID"))
        self.assertRC(response, 422)
        self.assertJE(
            response, 'read_time.@0',
            "Invalid format, it can be a number in hours, “1:05”, “:05”, "
            "“1h 5m”, “1.5h” or “30m”.")

    def test_not_null_field(self):
        author = Author.objects.create(
            first_name='Stephen',
            last_name='King',
            slug='stephen-king',
            create_comment=None)
        author.refresh_from_db()
        self.assertEqual(author.create_comment, "")



from django.conf import settings
from django.core.cache import cache
from django.utils import translation

from babel import Locale

from pfx.pfxcore.apidoc import Tag
from pfx.pfxcore.decorator import rest_api, rest_view
from pfx.pfxcore.http import JsonResponse

from .rest_views import BaseRestView

LOCALE_TAG = Tag("Locale")


def get_locales():
    return [
        dict(
            pk=code,
            name=Locale.parse(translation.to_locale(code)).get_display_name())
        for code, __ in settings.LANGUAGES]


def get_languages_json():
    locales = get_locales()
    return {
        'items': locales, 'meta': {'count': len(locales)}}


@rest_view("/locales")
class LocaleRestView(BaseRestView):
    """The view to get locale data."""
    tags = [LOCALE_TAG]

    @rest_api('/languages', public=True)
    def locales(self):
        """Entrypoint for :code:`GET /languages` route.

        Returns the list of available languages.

        :returns: The JSON response
        :rtype: :class:`JsonResponse`
        ---
        get:
            summary: Get languages
            responses:
                200:
                    content:
                        application/json:
                            schema:
                                properties:
                                    items:
                                        type: array
                                        items:
                                            type: object
                                            properties:
                                                pk:
                                                    type: string
                                                name:
                                                    type: string
                                    meta:
                                        type: object
                                        properties:
                                            count:
                                                type: number
        """
        data = cache.get_or_set('pfx.languages', get_languages_json, 60 * 60)
        return JsonResponse(data)



from django.core.exceptions import ValidationError
from django.http import JsonResponse as DjangoJsonResponse

from pfx.pfxcore.serializers.json import PFXJSONEncoder


class JsonResponse(DjangoJsonResponse):
    def __init__(self, data, encoder=PFXJSONEncoder, **kwargs):
        if isinstance(data, ValidationError):
            kwargs['safe'] = False
        super().__init__(data, encoder=encoder, **kwargs)



# Generated by Django 4.2.18 on 2025-01-20 13:24

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('pfxcore', '0002_pfxpermissionsuser'),
    ]

    operations = [
        migrations.DeleteModel(
            name='PFXPermissionsUser',
        ),
    ]



import logging

from django.conf import settings
from django.http import HttpResponseRedirect
from django.utils import translation
from django.utils.cache import patch_vary_headers
from django.utils.deprecation import MiddlewareMixin

import pytz

logger = logging.getLogger(__name__)


def get_language_from_request(request):
    """Custom version of django `get_language_from_request` from translation.

    Remove path and cookies parsing and just use `X-Custom-language` as
    first choice if it is defined and valid, then use `Accept-Language`.

    :param request: The HTTP request.
    :type request: HttpRequest
    :return: The language code.
    :rtype: str
    """
    _trans = translation._trans  # Load _trans dynamically from translation.

    x_custom_language = request.META.get('HTTP_X_CUSTOM_LANGUAGE', '')
    custom_lang = translation.to_language(x_custom_language)
    try:
        return _trans.get_supported_language_variant(custom_lang)
    except LookupError:
        logger.debug(
            "Unsupported x-custom-language header: "
            f"{x_custom_language} [lang={custom_lang}]")
        pass

    accept = request.META.get('HTTP_ACCEPT_LANGUAGE', '')
    for accept_lang, unused in _trans.parse_accept_lang_header(accept):
        if accept_lang == '*':
            break  # pragma: no cover
        if not _trans.language_code_re.search(accept_lang):
            continue  # pragma: no cover
        try:
            return _trans.get_supported_language_variant(accept_lang)
        except LookupError:
            logger.debug(
                f"Unsupported accept-language header: {accept}]")
            continue

    try:
        return _trans.get_supported_language_variant(settings.LANGUAGE_CODE)
    except LookupError:  # pragma: no cover
        logger.debug(
            f"Unsupported LANGUAGE_CODE: {settings.LANGUAGE_CODE}]")
        return settings.LANGUAGE_CODE


def get_timezone_from_request(request):
    """Get the timezone from request.

    If `X-Custom-Timezone` is defined in request headers and is valid (exists
    in `pytz.all_timezones_set`), it is returned. Otherwise, the default
    timezone (from DJango) is returned.

    :param request: The HTTP request.
    :type request: HttpRequest
    :return: The request timezone.
    :rtype: str
    """
    default_tz = getattr(settings, "TIME_ZONE", "UTC")
    tz = request.META.get('HTTP_X_CUSTOM_TIMEZONE', default_tz)
    if tz not in pytz.all_timezones_set:
        return default_tz
    return tz


class LocaleMiddleware(MiddlewareMixin):
    """A middleware to load locale data from request.

    * `request.LANGUAGE_CODE` is set from `get_language_from_request`.
    * `request.TIMEZONE` is set from `get_timezone_from_request`.
    * `Content-Language` is set in the response headers to specify the used
      language.
    """
    response_redirect_class = HttpResponseRedirect

    def process_request(self, request):
        language = get_language_from_request(request)
        translation.activate(language)
        request.LANGUAGE_CODE = translation.get_language()
        request.TIMEZONE = get_timezone_from_request(request)

    def process_response(self, request, response):
        patch_vary_headers(response, ('Accept-Language',))
        code = (
            len(request.LANGUAGE_CODE) == 5 and
            request.LANGUAGE_CODE[:3] + request.LANGUAGE_CODE[-2:].upper() or
            request.LANGUAGE_CODE)
        response.headers.setdefault('Content-Language', code)
        return response



from django.conf import settings
from django.core.management.commands.runserver import (
    Command as RunserverCommand,
)


class Command(RunserverCommand):
    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--keep-logging", dest="keep_logging", action='store_true')
        parser.add_argument(
            "--db-router", dest="db_router", action='store_true')
        parser.add_argument(
            "--sql", dest="sql_queries", default="", type=str)
        parser.add_argument(
            "--cprofile", dest="cprofile", default="", type=str)

    def run(self, *args, **options):
        """Runs the server with profile mode."""

        settings.DEBUG = True
        settings.MIDDLEWARE = [
            'pfx.pfxcore.middleware.ProfilingMiddleware',
        ] + settings.MIDDLEWARE
        if options['db_router']:
            settings.DATABASE_ROUTERS = [
                'pfx.pfxcore.middleware.ProfilingDatabaseRouter'
            ] + settings.DATABASE_ROUTERS
        settings.PFX_PROFILING_SQL = options['sql_queries']
        settings.PFX_PROFILING_CPROFILE = options['cprofile']
        if not options['keep_logging']:
            settings.LOGGING = {
                'version': 1,
                'disable_existing_loggers': False,
                'formatters': {
                    'console': {
                        'format': 'PROFILE {asctime} {message}',
                        'style': '{',
                    }
                },
                'handlers': {
                    'console': {
                        'class': 'logging.StreamHandler',
                        'formatter': 'console',
                    },
                },
                'root': {
                    'handlers': ['console'],
                    'level': 'ERROR',
                    'propagate': False,
                },
                'loggers': {
                    'django': {
                        'handlers': ['console'],
                        'level': 'ERROR',
                        'propagate': False,
                    },
                    'pfx.pfxcore.middleware.profiling': {
                        'handlers': ['console'],
                        'level': 'DEBUG',
                        'propagate': False,
                    }}}
        super(Command, self).run(*args, **options)



# Generated by Django 4.2.17 on 2025-01-21 09:43
# flake8: noqa

from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]



from datetime import date

from django.test import TestCase

from pfx.pfxcore.exceptions import APIError, ModelNotFoundAPIError
from pfx.pfxcore.shortcuts import (
    f,
    get_bool,
    get_date,
    get_float,
    get_int,
    get_object,
    get_pk,
    parse_bool,
    parse_date,
    parse_float,
    parse_int,
)
from pfx.pfxcore.test import TestAssertMixin
from tests.models import Author, Book


class ShortcutTest(TestAssertMixin, TestCase):

    def test_f(self):
        text = f('Test {first}, {second}', first='first', second='second')
        self.assertEqual(text, 'Test first, second')

    def test_get_object(self):
        with self.assertRaises(ModelNotFoundAPIError):
            get_object(Book.objects.all(), pk=-99)
        author = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        a = get_object(Author.objects.all(), pk=author.pk)
        self.assertEqual(a.pk, author.pk)

    def test_get_pk(self):
        author = dict(
            pk=122,
            resource_name='John Ronald Reuel Tolkien')
        pk = get_pk(122)
        self.assertEqual(pk, 122)
        pk = get_pk(author)
        self.assertEqual(pk, 122)

    def test_parse_bool(self):
        self.assertIsNone(parse_bool(None))
        self.assertIsNone(parse_bool(''))
        self.assertIsNone(parse_bool('null'))
        self.assertIsNone(parse_bool('undefined'))
        self.assertIs(parse_bool('false'), False)
        self.assertIs(parse_bool('False'), False)
        self.assertIs(parse_bool('FALSE'), False)
        self.assertIs(parse_bool('true'), True)
        self.assertIs(parse_bool('True'), True)
        self.assertIs(parse_bool('TRUE'), True)
        with self.assertRaises(ValueError):
            parse_bool('INVALID')

    def test_get_bool(self):
        vals = dict(
            none=None,
            empty='',
            null='null',
            undefined='undefined',
            false0='false',
            false1='False',
            false2='False',
            true0='true',
            true1='True',
            true2='True',
            invalid='INVALID')
        self.assertIsNone(get_bool(vals, 'none'))
        self.assertIsNone(get_bool(vals, 'empty'))
        self.assertIsNone(get_bool(vals, 'null'))
        self.assertIsNone(get_bool(vals, 'undefined'))
        self.assertIs(get_bool(vals, 'false0'), False)
        self.assertIs(get_bool(vals, 'false1'), False)
        self.assertIs(get_bool(vals, 'false2'), False)
        self.assertIs(get_bool(vals, 'true0'), True)
        self.assertIs(get_bool(vals, 'true1'), True)
        self.assertIs(get_bool(vals, 'true2'), True)
        with self.assertRaises(APIError):
            get_bool(vals, 'invalid')
        self.assertIs(get_bool(vals, 'does_not_exists'), None)
        self.assertIs(get_bool(vals, 'does_not_exists', default=True), True)

    def test_parse_int(self):
        self.assertIsNone(parse_int(None))
        self.assertIsNone(parse_int(''))
        self.assertIsNone(parse_int('null'))
        self.assertIsNone(parse_int('undefined'))
        self.assertEqual(parse_int('0'), 0)
        self.assertEqual(parse_int('-7'), -7)
        self.assertEqual(parse_int('13'), 13)
        with self.assertRaises(ValueError):
            parse_int('INVALID')

    def test_get_int(self):
        vals = dict(
            none=None,
            empty='',
            null='null',
            undefined='undefined',
            zero='0',
            negative='-7',
            positive='13',
            invalid='INVALID')
        self.assertIsNone(get_int(vals, 'none'))
        self.assertIsNone(get_int(vals, 'empty'))
        self.assertIsNone(get_int(vals, 'null'))
        self.assertIsNone(get_int(vals, 'undefined'))
        self.assertEqual(get_int(vals, 'zero'), 0)
        self.assertEqual(get_int(vals, 'negative'), -7)
        self.assertEqual(get_int(vals, 'positive'), 13)
        with self.assertRaises(APIError):
            get_int(vals, 'invalid')
        self.assertIs(get_int(vals, 'does_not_exists'), None)
        self.assertEqual(get_int(vals, 'does_not_exists', default=42), 42)

    def test_parse_float(self):
        self.assertIsNone(parse_float(None))
        self.assertIsNone(parse_float(''))
        self.assertIsNone(parse_float('null'))
        self.assertIsNone(parse_float('undefined'))
        self.assertEqual(parse_float('0'), 0)
        self.assertEqual(parse_float('3'), 3)
        self.assertEqual(parse_float('-7.45'), -7.45)
        self.assertEqual(parse_float('13.57'), 13.57)
        with self.assertRaises(ValueError):
            parse_float('INVALID')

    def test_get_float(self):
        vals = dict(
            none=None,
            empty='',
            null='null',
            undefined='undefined',
            zero='0',
            integer='3',
            negative='-7.45',
            positive='13.57',
            invalid='INVALID')
        self.assertIsNone(get_float(vals, 'none'))
        self.assertIsNone(get_float(vals, 'empty'))
        self.assertIsNone(get_float(vals, 'null'))
        self.assertIsNone(get_float(vals, 'undefined'))
        self.assertEqual(get_float(vals, 'zero'), 0)
        self.assertEqual(get_float(vals, 'integer'), 3)
        self.assertEqual(get_float(vals, 'negative'), -7.45)
        self.assertEqual(get_float(vals, 'positive'), 13.57)
        with self.assertRaises(APIError):
            get_float(vals, 'invalid')
        self.assertIs(get_float(vals, 'does_not_exists'), None)
        self.assertEqual(get_float(
            vals, 'does_not_exists', default=3.14159265359), 3.14159265359)

    def test_parse_date(self):
        self.assertIsNone(parse_date(None))
        self.assertIsNone(parse_date(''))
        self.assertIsNone(parse_date('null'))
        self.assertIsNone(parse_date('undefined'))
        self.assertEqual(parse_date('2020-05-17'), date(2020, 5, 17))
        with self.assertRaises(ValueError):
            parse_date('INVALID')

    def test_get_date(self):
        vals = dict(
            none=None,
            empty='',
            null='null',
            undefined='undefined',
            date='2020-05-17',
            invalid='INVALID')
        self.assertIsNone(get_date(vals, 'none'))
        self.assertIsNone(get_date(vals, 'empty'))
        self.assertIsNone(get_date(vals, 'null'))
        self.assertIsNone(get_date(vals, 'undefined'))
        self.assertEqual(get_date(vals, 'date'), date(2020, 5, 17))
        with self.assertRaises(APIError):
            get_date(vals, 'invalid')
        self.assertIs(get_date(vals, 'does_not_exists'), None)
        self.assertEqual(get_date(
            vals, 'does_not_exists', default=date(2020, 5, 21)
        ), date(2020, 5, 21))



from django.contrib.auth.models import BaseUserManager
from django.db import models

from pfx.pfxcore.models import AbstractPFXBaseUser


class User(AbstractPFXBaseUser):
    """Default user for tests."""

    USERNAME_FIELD = 'username'

    username = models.CharField("Username", max_length=150, unique=True)

    objects = BaseUserManager()

    class Meta:
        verbose_name = 'user'
        verbose_name_plural = 'users'



import logging
import re
from datetime import timedelta

from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.signals import post_delete
from django.dispatch import receiver
from django.utils.translation import gettext_lazy as _

from pfx.pfxcore.storage import DefaultStorage

logger = logging.getLogger(__name__)


class MediaField(models.JSONField):
    def __init__(
            self, *args, max_length=255, get_key=None, storage=None,
            auto_delete=False, **kwargs):
        self.get_key = get_key or self.get_default_key
        self.storage = storage or DefaultStorage()
        self.auto_delete = auto_delete
        super().__init__(
            *args, max_length=max_length,
            default=kwargs.pop('default', dict),
            blank=kwargs.pop('blank', True),
            **kwargs)

    @staticmethod
    def get_default_key(obj, filename):
        return f"{type(obj).__name__}/{obj.pk}/{filename}"

    def to_python(self, value):
        return super().to_python(self.storage.to_python(value))

    def get_upload_url(self, request, obj, filename):
        key = self.get_key(obj, filename)
        url = self.storage.get_upload_url(request, key)
        return dict(url=url, file=dict(name=filename, key=key))

    def get_url(self, request, obj):
        return self.storage.get_url(
            request, self.value_from_object(obj)['key'])

    def upload(self, obj, file, filename, **kwargs):
        key = self.get_key(obj, filename)
        self.storage.upload(key, file, **kwargs)
        return self.to_python(dict(key=key))


@receiver(post_delete)
def post_delete_media(sender, instance, **kwargs):
    for field in sender._meta.fields:
        if isinstance(field, MediaField) and field.auto_delete:
            field.storage.delete(field.value_from_object(instance))


class MinutesDurationField(models.DurationField):
    RE_FLOAT = re.compile(r'^[0-9]*(\.[0-9]*)?$')
    RE_HH_MM = re.compile(r'^([0-9]*):([0-5][0-9])?$')
    RE_HUMAN = re.compile(
        r'^\s*(?:([0-9]*(?:\.[0-9]*)?)h)?\s*(?:([0-9]*)m)?\s*$')
    schema = dict(type='object', properties=dict(
        minutes=dict(type='number', example=90),
        clock_format=dict(type='string', example='01:30'),
        human_format=dict(type='string', example='1h 30m')))

    def to_python(self, value):
        if value is None or value == '':
            return None
        if isinstance(value, timedelta):
            return value
        if isinstance(value, (int, float)):
            return timedelta(hours=value)
        if not isinstance(value, str):
            logger.error(f"invalid value {value} [{type(value)}]")
            raise ValidationError(_("Invalid value."))
        match_float = self.RE_FLOAT.match(value)
        if match_float:
            return timedelta(hours=float(value))
        match_hm = self.RE_HH_MM.match(value)
        if match_hm:
            h, m = match_hm.groups()
            return timedelta(
                hours=h and int(h) or 0, minutes=m and int(m) or 0)
        match_human = self.RE_HUMAN.match(value)
        if match_human:
            h, m = match_human.groups()
            return timedelta(
                hours=h and float(h) or 0, minutes=m and int(m) or 0)
        raise ValidationError(_(
            "Invalid format, it can be a number in hours, “1:05”, “:05”, "
            "“1h 5m”, “1.5h” or “30m”."))

    @staticmethod
    def to_json(value):
        if value is None:
            return None
        minutes = int(value.total_seconds() / 60)
        h, m = minutes // 60, minutes % 60
        return dict(
            minutes=minutes,
            clock_format=f"{minutes // 60}:{minutes % 60:02d}",
            human_format=(
                f'{h and f"{h}h" or ""}\u00A0'
                f'{m and f"{m}m" or ""}'.strip()))




from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.utils.translation import gettext_lazy as _

from pfx.pfxcore.exceptions import APIError
from pfx.pfxcore.models import OrderedModelMixin
from pfx.pfxcore.shortcuts import f

from .rest_views import ModelResponseMixin


class OrderedRestViewMixin(ModelResponseMixin):
    """Extension mixin for OrderedModelMixin models."""

    def __init__(self, *args, **kwargs):
        if not issubclass(self.model, OrderedModelMixin):
            raise TypeError("model is not a subclass of OrderedModelMixin")
        super().__init__(*args, **kwargs)

    def is_valid(self, obj, created=False, **kwargs):
        with transaction.atomic():
            res = super().is_valid(obj, created, **kwargs)
            self.apply_move(obj)
        return res

    def apply_move(self, obj):
        def get_object(move):
            pk = self.request.GET.get('object')
            if not pk:
                raise APIError(f(
                    _("object parameter is mandatory for move={move}."),
                    move=move))
            try:
                return self.get_queryset(
                    from_queryset=obj._order_qs).get(pk=pk)
            except ObjectDoesNotExist:
                raise APIError(f(
                    _("object {pk} does not exists in this move context."),
                    pk=pk))

        move = self.request.GET.get('move')
        if move == 'up':
            obj.up()
        elif move == 'down':
            obj.down()
        elif move == 'bottom':
            obj.bottom()
        elif move == 'top':
            obj.top()
        elif move == 'above':
            obj.above(get_object(move))
        elif move == 'below':
            obj.below(get_object(move))



from .parameters import (
    Parameter,
    ParameterGroup,
    QueryParameter,
    register_global_parameter,
)
from .schema import ModelListSchema, ModelSchema, Schema
from .tags import Tag

__PARAMETERS__ = []
__TAGS__ = {}



from functools import reduce

from django.db.models import Q
from django.utils.translation import gettext_lazy as _

from pfx.pfxcore.exceptions import APIError
from pfx.pfxcore.shortcuts import (
    f,
    parse_bool,
    parse_date,
    parse_float,
    parse_int,
)
from pfx.pfxcore.views import FieldType


class FilterGroup():
    def __init__(self, name, label, filters):
        self.name = name
        self.label = label
        self.filters = filters

    @property
    def meta(self):
        return dict(is_group=True, name=self.name, label=self.label, items=[
            f.meta for f in self.filters
        ])

    def query(self, params):
        return Q(*[f.query(params) for f in self.filters])


class Filter():
    def __init__(
            self, name, label, type=None, filter_func=None,
            filter_func_and=False, filter_func_list=False, choices=None,
            related_model=None, related_model_api=None,
            technical=False, defaults=None, empty_value=True):
        self.name = name
        self.label = label
        self.type = type
        self.filter_func = filter_func
        self.filter_func_and = filter_func_and
        self.filter_func_list = filter_func_list
        self.choices = choices
        self.related_model = related_model
        self.related_model_api = related_model_api
        self.technical = technical
        self.defaults = defaults or []
        self.empty_value = empty_value

    @property
    def meta(self):
        res = dict(
            is_group=False,
            label=_(self.label),
            name=self.name,
            type=self.type,
            empty_value=self.empty_value,
            technical=self.technical)
        if self.choices:
            res['choices'] = [
                dict(label=_(v), value=k) for k, v in self.choices]
        if self.related_model:
            res['related_model'] = str(self.related_model.__name__)
            res['api'] = self.related_model_api
        return res

    def _parse_value(self, value):
        try:
            if self.type == FieldType.BooleanField:
                return parse_bool(value)
            if self.type in (FieldType.IntegerField, FieldType.ModelObject):
                return parse_int(value)
            if self.type == FieldType.FloatField:
                return parse_float(value)
            if self.type == FieldType.DateField:
                return parse_date(value)
        except ValueError:
            raise APIError(f(
                _("Invalid value for {filter} filter"), filter=self.label))
        return value

    def _call_filter_func(self, values):
        if self.filter_func_list:
            return self.filter_func(values)
        return reduce(
            lambda x, y: x & y if self.filter_func_and else x | y,
            [self.filter_func(v) for v in values])

    def _get_values(self, params):
        values = []
        if (self.type == FieldType.ModelObject and self.related_model and
                hasattr(self.related_model.objects, 'default_search')):
            q = None
            for search in params.getlist(f'{self.name}*'):
                crit = self.related_model.objects.default_search(search)
                if q is None:
                    q = crit
                q |= crit
            if q:
                values.extend(self.related_model.objects.filter(
                    q).values_list('pk', flat=True) or [-1])
        values.extend([
            self._parse_value(v) for v in params.getlist(self.name)])
        return values or self.defaults

    def query(self, params):
        values = self._get_values(params)
        return self._call_filter_func(values) if values else Q()


class ModelFilter(Filter):
    def __init__(
            self, model, name, label=None, type=None,
            filter_func=None, filter_func_and=False, filter_func_list=False,
            choices=None, related_model=None, related_model_api=None,
            technical=False, defaults=None, empty_value=None):
        self.model = model
        self.field = model._meta.get_field(name)
        if empty_value is None:
            empty_value = self.field.blank
        super().__init__(
            name, label or self.field.verbose_name,
            type or self._type_from_model,
            filter_func, filter_func_and, filter_func_list,
            choices or self.field.choices,
            related_model or (
                self.field.remote_field and
                self.field.remote_field.model),
            related_model_api or (
                self.field.remote_field and
                self.field.remote_field.model and
                hasattr(self.field.remote_field.model, 'api') and
                self.field.remote_field.model.api),
            technical=technical, defaults=defaults, empty_value=empty_value)

    @property
    def _type_from_model(self):
        model_type = FieldType.from_model_field(self.field.__class__)
        if model_type == FieldType.ModelObjectList:
            # For list related fields (OneToMany, ManyToMany, …) the field
            # type is ModelObjectList, but the filter type should be
            # ModelObject.
            return FieldType.ModelObject
        return model_type

    def query(self, params):
        values = self._get_values(params)
        if self.filter_func and values:
            return self._call_filter_func(values)
        elif values:
            return reduce(
                lambda x, y: x & y if self.filter_func_and else x | y,
                [Q(**{self.name: v}) for v in values])
        return Q()




import logging
from datetime import date

from django.core.exceptions import ObjectDoesNotExist
from django.utils.translation import gettext_lazy as _

from .settings import settings

logger = logging.getLogger(__name__)


def f(tmpl, **kwargs):
    return tmpl.format(**kwargs)


def get_object(queryset, related_field=None, **kwargs):
    from .exceptions import ModelNotFoundAPIError, RelatedModelNotFoundAPIError
    try:
        return queryset.get(**kwargs)
    except ObjectDoesNotExist:
        if related_field:
            raise RelatedModelNotFoundAPIError(related_field, queryset.model)
        raise ModelNotFoundAPIError(queryset.model)


def get_pk(obj):
    if isinstance(obj, dict) and 'pk' in obj:
        return obj['pk']
    return obj


def is_null(value):
    return not value or value.lower() in ('null', 'undefined')


def parse_int(value):
    if is_null(value):
        return None
    return int(value)


def get_int(data, key, default=None):
    if key not in data:
        return default
    try:
        return parse_int(data.get(key))
    except ValueError:
        from pfx.pfxcore.exceptions import APIError
        raise APIError(f(_("{key} must be an integer number."), key=key))


def parse_float(value):
    if is_null(value):
        return None
    return float(value)


def get_float(data, key, default=None):
    if key not in data:
        return default
    try:
        return parse_float(data.get(key))
    except ValueError:
        from pfx.pfxcore.exceptions import APIError
        raise APIError(f(_("{key} must be a number."), key=key))


def parse_date(value):
    if is_null(value):
        return None
    return date.fromisoformat(value)


def get_date(data, key, default=None):
    if key not in data:
        return default
    try:
        return parse_date(data.get(key))
    except ValueError:
        from pfx.pfxcore.exceptions import APIError
        raise APIError(f(_("{key} must be a date."), key=key))


def parse_bool(value):
    if is_null(value):
        return None
    if value.lower() in ('true', '1'):
        return True
    if value.lower() in ('false', '0'):
        return False
    raise ValueError(f"{value} is not a valid boolean value.")


def get_bool(data, key, default=None):
    if key not in data:
        return default
    try:
        return parse_bool(data.get(key))
    except ValueError:
        from pfx.pfxcore.exceptions import APIError
        raise APIError(f(
            _("{key} must be “true”, “false”, “1”, “0” or empty."),
            key=key))


def delete_token_cookie(response):
    response.delete_cookie(
        'token',
        domain=settings.PFX_COOKIE_DOMAIN,
        samesite=settings.PFX_COOKIE_SAMESITE)
    return response


def register_views(*views):
    def _register():
        for v in views:
            yield from v.as_urlpatterns()
    return list(_register())


def class_key(cls, *args):
    return f"{cls.__module__}.{cls.__name__}{''.join(f'.{a}' for a in args)}"


def permissions(*perms):
    from django.contrib.auth.models import Permission
    pks = set()
    for perm in perms:
        app_label, codename = perm.split('.')
        try:
            pks.add(Permission.objects.get(
                codename=codename, content_type__app_label=app_label).pk)
        except Permission.DoesNotExist:
            raise Exception(f"Permission {perm} does not exists.")
    return Permission.objects.filter(pk__in=pks)


def model_permissions(model, *actions):
    meta = model._meta
    return {f"{meta.app_label}.{a}_{meta.model_name}" for a in actions}



from .authentication import (
    AuthenticationMiddleware,
    CookieAuthenticationMiddleware,
)
from .locale import LocaleMiddleware
from .profiling import ProfilingDatabaseRouter, ProfilingMiddleware



from django.db.migrations.operations.base import Operation


class GroupPermsOperation(Operation):
    def state_forwards(self, app_label, state):
        pass


class CreateGroup(GroupPermsOperation):
    """Create a group."""

    def __init__(self, name):
        self.group_name = name

    def deconstruct(self):
        return (self.__class__.__qualname__, [self.group_name], {})

    def database_forwards(
            self, app_label, schema_editor, from_state, to_state):
        Group = to_state.apps.get_model('auth', 'Group')
        if self.allow_migrate_model(schema_editor.connection.alias, Group):
            Group.objects.create(name=self.group_name)

    def database_backwards(
            self, app_label, schema_editor, from_state, to_state):
        Group = to_state.apps.get_model('auth', 'Group')
        if self.allow_migrate_model(schema_editor.connection.alias, Group):
            Group.objects.get(name=self.group_name).delete()

    def describe(self):
        return (f"Create group {self.group_name}")


class DeleteGroup(GroupPermsOperation):
    """Delete a group."""

    def __init__(self, name):
        self.group_name = name

    def deconstruct(self):
        return (self.__class__.__qualname__, [self.group_name], {})

    def database_forwards(
            self, app_label, schema_editor, from_state, to_state):
        Group = to_state.apps.get_model('auth', 'Group')
        if self.allow_migrate_model(schema_editor.connection.alias, Group):
            Group.objects.get(name=self.group_name).delete()

    def database_backwards(
            self, app_label, schema_editor, from_state, to_state):
        Group = to_state.apps.get_model('auth', 'Group')
        if self.allow_migrate_model(schema_editor.connection.alias, Group):
            Group.objects.create(name=self.group_name)

    def describe(self):
        return (f"Create group {self.group_name}")


class RenameGroup(GroupPermsOperation):
    """Rename a group."""

    def __init__(self, old_name, new_name):
        self.group_old_name = old_name
        self.group_new_name = new_name

    def deconstruct(self):
        return (self.__class__.__qualname__, [
            self.group_old_name, self.group_new_name], {})

    def database_forwards(
            self, app_label, schema_editor, from_state, to_state):
        Group = to_state.apps.get_model('auth', 'Group')
        if self.allow_migrate_model(schema_editor.connection.alias, Group):
            group = Group.objects.get(name=self.group_old_name)
            group.name = self.group_new_name
            group.save()

    def database_backwards(
            self, app_label, schema_editor, from_state, to_state):
        Group = to_state.apps.get_model('auth', 'Group')
        if self.allow_migrate_model(schema_editor.connection.alias, Group):
            group = Group.objects.get(name=self.group_new_name)
            group.name = self.group_old_name
            group.save()

    def describe(self):
        return (f"Rename group {self.group_old_name} to {self.group_new_name}")



import os

BASE_DIR = os.path.dirname(os.path.abspath(__file__))

SECRET_KEY = 'fake-key'
INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.postgres',
    'pfx.pfxcore',
    'tests_base_user',
]

MIDDLEWARE = [
    'pfx.pfxcore.middleware.LocaleMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.middleware.common.CommonMiddleware',
    'pfx.pfxcore.middleware.AuthenticationMiddleware',
    'pfx.pfxcore.middleware.CookieAuthenticationMiddleware',
]

USE_I18N = True
USE_L10N = True
LANGUAGE_CODE = 'en'
LANGUAGES = [
    ('en', 'English'),
    ('fr', 'French')]
LOCALE_PATHS = [
    os.path.join(BASE_DIR, "tests_base_user/locale")]

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

AUTH_USER_MODEL = 'pfxcore.PFXUser'

PFX_SECRET_KEY = "fake-secret-key"
PFX_COOKIE_DOMAIN = None

PFX_MAX_LIST_RESULT_SIZE = 0

ROOT_URLCONF = 'tests_base_user.urls'
APPEND_SLASH = False

EMAIL_BACKEND = 'django.core.mail.backends.console.EmailBackend'
PFX_RESET_PASSWORD_URL = (
    'http://localhost:8000/test?token={token}&uidb64={uidb64}')
PFX_SITE_NAME = 'Books Demo'

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
    }
]

LOGGING = {
    'version': 1,
    'disable_existing_logger': False,
    'formatters': {
        'console': {
            'format': "\n%(name)-25s %(levelname)-8s %(message)s",
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'console',
        },
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'ERROR',
            'propagate': True,
        },
    },
}



# Generated by Django 4.2.17 on 2025-01-22 07:34
# flake8: noqa

import django.contrib.auth.models
import django.contrib.auth.validators
import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models

import pfx.pfxcore.fields
import pfx.pfxcore.models.cache_mixins
import pfx.pfxcore.models.not_null_fields
import pfx.pfxcore.models.pfx_models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0012_alter_user_first_name_max_length'),
    ]

    operations = [
        migrations.CreateModel(
            name='Author',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('first_name', models.CharField(max_length=30, verbose_name='First Name')),
                ('last_name', models.CharField(max_length=30, verbose_name='Last Name')),
                ('slug', models.SlugField(unique=True, verbose_name='Slug')),
                ('gender', models.CharField(choices=[('male', 'Male'), ('female', 'Female')], default='male', max_length=10, verbose_name='Gender')),
                ('science_fiction', models.BooleanField(default=False, verbose_name='Science Fiction')),
                ('created_at', models.DateField(auto_now_add=True, verbose_name='Created at')),
                ('create_comment', pfx.pfxcore.models.not_null_fields.NotNullCharField(blank=True, max_length=255, verbose_name='Create comment')),
                ('update_comment', pfx.pfxcore.models.not_null_fields.NotNullCharField(blank=True, max_length=255, verbose_name='Update comment')),
                ('website', pfx.pfxcore.models.not_null_fields.NotNullURLField(blank=True, max_length=255, verbose_name='Website')),
            ],
            options={
                'verbose_name': 'Author',
                'verbose_name_plural': 'Authors',
                'ordering': ['last_name', 'first_name', 'pk'],
                'permissions': [('can_customize_author', 'Can customize author')],
            },
            bases=(pfx.pfxcore.models.cache_mixins.CacheableMixin, pfx.pfxcore.models.pfx_models.JSONReprMixin, models.Model),
        ),
        migrations.CreateModel(
            name='BookType',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=30, verbose_name='Name')),
                ('slug', models.SlugField(verbose_name='Slug')),
            ],
            options={
                'verbose_name': 'Book Type',
                'verbose_name_plural': 'Book Types',
            },
            bases=(pfx.pfxcore.models.cache_mixins.CacheDependsMixin, pfx.pfxcore.models.pfx_models.PFXModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='Book',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=100, verbose_name='Name')),
                ('pub_date', models.DateField(verbose_name='Pub Date')),
                ('created_at', models.DateField(auto_now_add=True, verbose_name='Created at')),
                ('pages', models.IntegerField(blank=True, null=True, verbose_name='Pages')),
                ('rating', models.FloatField(blank=True, null=True, verbose_name='Rating')),
                ('reference', models.CharField(blank=True, max_length=30, null=True, verbose_name='Reference')),
                ('cover', pfx.pfxcore.fields.MediaField(blank=True, default=dict, max_length=255, verbose_name='Cover')),
                ('read_time', pfx.pfxcore.fields.MinutesDurationField(blank=True, null=True, verbose_name='Read Time')),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.RESTRICT, related_name='books', to='tests.author', verbose_name='Author')),
                ('type', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='books', to='tests.booktype', verbose_name='Book Type')),
            ],
            options={
                'verbose_name': 'Book',
                'verbose_name_plural': 'Books',
            },
            bases=(pfx.pfxcore.models.cache_mixins.CacheDependsMixin, pfx.pfxcore.models.pfx_models.PFXModelMixin, models.Model),
        ),
        migrations.AddField(
            model_name='author',
            name='types',
            field=models.ManyToManyField(related_name='authors', to='tests.booktype', verbose_name='Types'),
        ),
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(max_length=128, verbose_name='password')),
                ('last_login', models.DateTimeField(blank=True, null=True, verbose_name='last login')),
                ('is_superuser', models.BooleanField(default=False, help_text='Designates that this user has all permissions without explicitly assigning them.', verbose_name='superuser status')),
                ('username', models.CharField(error_messages={'unique': 'A user with that username already exists.'}, help_text='Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.', max_length=150, unique=True, validators=[django.contrib.auth.validators.UnicodeUsernameValidator()], verbose_name='username')),
                ('first_name', models.CharField(blank=True, max_length=150, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=150, verbose_name='last name')),
                ('email', models.EmailField(blank=True, max_length=254, verbose_name='email address')),
                ('is_staff', models.BooleanField(default=False, help_text='Designates whether the user can log into this admin site.', verbose_name='staff status')),
                ('is_active', models.BooleanField(default=True, help_text='Designates whether this user should be treated as active. Unselect this instead of deleting accounts.', verbose_name='active')),
                ('date_joined', models.DateTimeField(default=django.utils.timezone.now, verbose_name='date joined')),
                ('otp_secret_token', models.CharField(blank=True, max_length=32, null=True, unique=True, verbose_name='OTP secret token')),
                ('otp_secret_token_tmp', models.CharField(blank=True, max_length=32, null=True, verbose_name='Temporary OTP secret token')),
                ('hotp_count', models.IntegerField(default=0, verbose_name='HOTP count')),
                ('hotp_expiry', models.DateTimeField(default=django.utils.timezone.now, verbose_name='HOTP expiry')),
                ('groups', models.ManyToManyField(blank=True, help_text='The groups this user belongs to. A user will get all permissions granted to each of their groups.', related_name='user_set', related_query_name='user', to='auth.group', verbose_name='groups')),
                ('user_permissions', models.ManyToManyField(blank=True, help_text='Specific permissions for this user.', related_name='user_set', related_query_name='user', to='auth.permission', verbose_name='user permissions')),
            ],
            options={
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
            },
            bases=(pfx.pfxcore.models.cache_mixins.CacheableMixin, pfx.pfxcore.models.pfx_models.PFXModelMixin, models.Model),
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.AddConstraint(
            model_name='book',
            constraint=pfx.pfxcore.models.pfx_models.UniqueConstraint(fields=('author', 'name'), message='%(name)s already exists for %(author)s', name='book_unique_author_and_name'),
        ),
    ]



import inspect
import logging
import operator
from datetime import date

from django.db import models
from django.db.models.constants import LOOKUP_SEP
from django.utils.formats import date_format
from django.utils.functional import cached_property

from apispec.utils import deepupdate
from apispec.yaml_utils import load_yaml_from_docstring

from pfx.pfxcore import fields as pfx_fields
from pfx.pfxcore.models import JSONReprMixin
from pfx.pfxcore.shortcuts import get_object

logger = logging.getLogger(__name__)


def setifnone(kwargs, k, default):
    if kwargs.get(k) is None:
        kwargs[k] = default


class ModelList(list):
    pass


class FieldType:
    CharField = "CharField"
    TextField = "TextField"
    BooleanField = "BooleanField"
    IntegerField = "IntegerField"
    FloatField = "FloatField"
    DecimalField = "DecimalField"
    DateField = "DateField"
    DateTimeField = "DateTimeField"
    MinutesDurationField = "MinutesDurationField"
    MediaField = "MediaField"
    ModelObject = "ModelObject"
    ModelObjectList = "ModelObjectList"
    JsonObject = "JsonObject"

    MODEL_FIELD_BINDING = [
        (pfx_fields.MinutesDurationField, MinutesDurationField),
        (models.BooleanField, BooleanField),
        (models.IntegerField, IntegerField),
        (models.FloatField, FloatField),
        (models.DecimalField, DecimalField),
        (models.DateTimeField, DateTimeField),
        (models.DateField, DateField),
        (models.TextField, TextField),
        (models.CharField, CharField),
        (models.URLField, CharField),
        (models.UUIDField, CharField),
        (models.ForeignObject, ModelObject),
        (models.ForeignObjectRel, ModelObjectList),
        (models.ManyToManyField, ModelObjectList),
        (models.JSONField, JsonObject),
    ]
    APIDOC_FIELD_BINDING = {
        CharField: "string",
        TextField: "string",
        BooleanField: "boolean",
        IntegerField: "number",
        FloatField: "number",
        DecimalField: "string",
        DateField: "string",
        DateTimeField: "string",
        MinutesDurationField: "object",
        MediaField: "object",
        ModelObject: "object",
        ModelObjectList: "array",
        JsonObject: "object",
    }

    @classmethod
    def register_binding(cls, field_class, field_type, apidoc_type='object'):
        cls.MODEL_FIELD_BINDING.insert(0, (field_class, field_type))
        cls.APIDOC_FIELD_BINDING[field_type] = apidoc_type

    @classmethod
    def from_model_field(cls, field_class):
        for k, v in FieldType.MODEL_FIELD_BINDING:
            if issubclass(field_class, k):
                return v

    @classmethod
    def to_apidoc(cls, field_type):
        return cls.APIDOC_FIELD_BINDING.get(field_type)


class ViewField:
    def __init__(
            self, name, verbose_name=None, alias=None, field_type=None,
            readonly=False, readonly_create=False, readonly_update=False,
            choices=None, json_repr=None,
            related_model=None, related_model_api=None,
            select_related=None, prefetch_related=None):
        self.name = name
        self.alias = alias or name
        self.readonly_create = readonly or readonly_create
        self.readonly_update = readonly or readonly_update
        self.verbose_name = verbose_name or name
        self.field_type = field_type
        self.choices = dict(choices or [])
        self.json_repr = json_repr
        self.related_model = related_model
        self.related_model_api = related_model_api
        self.select_related = select_related or []
        self.prefetch_related = prefetch_related or []

    def is_readonly(self, created=False):
        return self.readonly_create if created else self.readonly_update

    def meta(self):
        res = dict(type=self.field_type, name=self.verbose_name)
        if self.choices:
            res['choices'] = [
                dict(label=str(v), value=k) for k, v in self.choices.items()]
        if self.related_model:
            res['model'] = self.related_model._meta.label
            res['api'] = self.related_model_api or getattr(
                self.related_model, 'api', None)
        res['readonly'] = dict(
            post=self.readonly_create,
            put=self.readonly_update)
        return res

    def get_value(self, obj):
        dotstr = self.name.replace(LOOKUP_SEP, '.')
        if '.' in dotstr:
            path, name = dotstr.rsplit('.', 1)
            obj = operator.attrgetter(path)(obj)
        else:
            name = self.name
        return getattr(obj, name)

    def _json_repr(self, value):
        if not value:
            return None
        if self.json_repr:
            return self.json_repr(value)
        return value.json_repr()

    def to_json(self, obj, format_date):
        value = self.get_value(obj)
        if format_date and isinstance(value, date):
            return dict(
                value=value,
                formatted=date_format(
                    value, format='SHORT_DATE_FORMAT', use_l10n=True))
        if self.field_type == FieldType.MinutesDurationField:
            return pfx_fields.MinutesDurationField.to_json(value)
        if self.field_type == FieldType.ModelObject:
            return self._json_repr(value)
        if self.field_type == FieldType.ModelObjectList:
            return [self._json_repr(o) for o in value.all()]
        if self.choices:
            if value in self.choices:
                return dict(value=value, label=str(self.choices[value]))
            else:
                return None
        return value

    @classmethod
    def from_property(cls, name, prop, **kwargs):
        if hasattr(prop, 'fget'):
            verbose_name = getattr(
                prop.fget, 'short_description', prop.fget.__name__)
            field_type = getattr(prop.fget, 'field_type', None)
        else:
            verbose_name = (
                hasattr(prop, 'name') and prop.name or str(prop))
            field_type = None
        setifnone(kwargs, 'verbose_name', verbose_name)
        setifnone(kwargs, 'field_type', field_type)
        kwargs['readonly'] = True
        return ViewField(name, **kwargs)

    @classmethod
    def from_model_field(cls, name, field, **kwargs):
        setifnone(kwargs, 'verbose_name', cls._get_model_verbose_name(field))
        setifnone(
            kwargs, 'field_type', FieldType.from_model_field(field.__class__))
        return ViewModelField(
            name, model_field=field, **kwargs)

    @classmethod
    def from_name(cls, model, name, **kwargs):
        attr_model, attr_name = cls._resolve_lookup(model, name)
        try:
            attr = getattr(attr_model, attr_name)
        except AttributeError:
            kwargs['readonly'] = True
            return ViewField(name, **kwargs)
        if isinstance(attr, (property, cached_property)):
            return cls.from_property(name, attr, **kwargs)
        field = attr_model._meta.get_field(attr_name)
        if LOOKUP_SEP in name:
            kwargs['readonly'] = True
        return cls.from_model_field(name, field, **kwargs)

    @classmethod
    def _resolve_lookup(cls, model, name):
        path = name.split(LOOKUP_SEP)
        path, name = path[:-1], path[-1]
        for e in path:
            model = model._meta.get_field(e).related_model
        return model, name

    @classmethod
    def _get_model_verbose_name(cls, field):
        if hasattr(field, 'verbose_name'):
            return field.verbose_name
        elif hasattr(field, 'related_model'):
            if (hasattr(field, 'multiple') and field.multiple and
                    hasattr(field.related_model._meta, 'verbose_name_plural')):
                return field.related_model._meta.verbose_name_plural
            elif hasattr(field.related_model._meta, 'verbose_name'):
                return field.related_model._meta.verbose_name
        return field.name

    def to_apidoc(self, request=False):
        res = dict(type=FieldType.to_apidoc(self.field_type))
        if self.field_type == FieldType.DateField:
            res['format'] = 'date'
        elif self.field_type == FieldType.DateTimeField:
            res['format'] = 'date-time'
        elif self.field_type == FieldType.ModelObject:
            properties = {}
            if self.related_model:
                res['format'] = str(self.related_model._meta.verbose_name)
                doc = None
                if issubclass(self.related_model, JSONReprMixin):
                    properties.update(self.related_model.json_repr_schema())
                    doc = inspect.getdoc(self.related_model.json_repr)
                if doc:
                    properties = deepupdate(
                        load_yaml_from_docstring(doc), properties)
            if self.json_repr:
                doc = inspect.getdoc(self.json_repr)
                if doc:
                    p = load_yaml_from_docstring(doc)
                    if p.pop('_extends', False):
                        properties = deepupdate(p, properties)
                    else:
                        properties = p
            if properties:
                if request:
                    res.pop('type', None)
                    res.pop('format', None)
                    res['oneOf'] = [
                        properties.get('pk'),
                        dict(type='object',
                             properties=dict(pk=properties.get('pk')))]
                else:
                    res['properties'] = properties
        elif self.field_type == FieldType.MinutesDurationField:
            res = pfx_fields.MinutesDurationField.schema
        elif self.field_type == FieldType.ModelObjectList:
            res['items'] = dict(type='object')
            properties = {}
            if self.related_model:
                res['items']['format'] = str(
                    self.related_model._meta.verbose_name)
                doc = None
                if issubclass(self.related_model, JSONReprMixin):
                    properties.update(self.related_model.json_repr_schema())
                    doc = inspect.getdoc(self.related_model.json_repr)
                if doc:
                    properties = deepupdate(
                        load_yaml_from_docstring(doc), properties)
            if self.json_repr:
                doc = inspect.getdoc(self.json_repr)
                if doc:
                    p = load_yaml_from_docstring(doc)
                    if p.pop('_extends', False):
                        properties = deepupdate(p, properties)
                    else:
                        properties = p
            if properties:
                if request:
                    res.pop('type', None)
                    res.pop('format', None)
                    res['oneOf'] = [
                        properties.get('pk'),
                        dict(type='object',
                             properties=dict(pk=properties.get('pk')))]
                else:
                    res['items']['properties'] = properties
        res['readonly'] = self.readonly_create and self.readonly_update
        if self.choices:
            res['enum'] = list(self.choices)
        return res


class ViewModelField(ViewField):
    def __init__(
            self, name, model_field=None, **kwargs):
        super().__init__(name, **kwargs)
        self.model_field = model_field
        if (hasattr(self.model_field, 'related_model') and
                self.model_field.related_model and not self.related_model):
            self.related_model = self.model_field.related_model
        if not self.select_related and (
                self.field_type == FieldType.ModelObject):
            # Auto add the field in select_related if select_related
            # is not defined and the field is an object.
            self.select_related = [self.name]
        if not self.prefetch_related and (
                self.field_type == FieldType.ModelObjectList):
            # Auto add the field in prefetch_related if prefetch_related
            # is not defined and the field is an object list.
            self.prefetch_related = [self.name]
        self.choices = self.choices or dict(
            hasattr(model_field, 'choices') and model_field.choices or [])
        if not self.field_type:
            self.readonly = True

    def meta(self):
        res = super().meta()
        res['required'] = not (
            getattr(self.model_field, 'null', False) or
            getattr(self.model_field, 'blank', False))
        return res

    def to_model_value(self, value, get_related_queryset):
        def _get_obj(v):
            pk = v['pk'] if isinstance(v, dict) and 'pk' in v else v
            return pk and get_object(
                get_related_queryset(self.model_field.related_model),
                related_field=self.name, pk=pk) or None

        field = self.model_field
        if self.field_type == FieldType.ModelObject:
            value = _get_obj(value)
        elif self.field_type == FieldType.ModelObjectList:
            value = ModelList(_get_obj(v) for v in value)
        elif hasattr(field, 'choices') and field.choices:
            value = (value['value']
                     if isinstance(value, dict) and 'value' in value
                     else value)
        elif self.field_type == FieldType.DateField:
            value = (value['value']
                     if isinstance(value, dict) and 'value' in value
                     else value)
        elif self.field_type == FieldType.IntegerField:
            value = value if value != '' else None
        elif self.field_type == FieldType.FloatField:
            value = value if value != '' else None
        elif self.field_type == FieldType.MinutesDurationField:
            if value is None:
                pass
            elif value == '':
                value = None
            elif (isinstance(value, dict) and 'human_format' in value):
                value = value['human_format']
        return field.name, value

    def to_apidoc(self, request=False):
        res = super().to_apidoc(request)
        if self.model_field.null:
            res['nullable'] = True
        return res


class VF:
    def __init__(
            self, name, verbose_name=None, field_type=None, alias=None,
            readonly=False, readonly_create=False, readonly_update=False,
            choices=None, select_related=None, prefetch_related=None,
            json_repr=None, related_model=None, related_model_api=None):
        self.kwargs = dict(
            name=name, verbose_name=verbose_name, field_type=field_type,
            alias=alias,
            readonly=readonly, readonly_create=readonly_create,
            readonly_update=readonly_update, choices=choices,
            select_related=select_related, prefetch_related=prefetch_related,
            json_repr=json_repr, related_model=related_model,
            related_model_api=related_model_api)

    def to_field(self, model):
        return ViewField.from_name(model, **self.kwargs)



from pfx.pfxcore.apidoc import ParameterGroup

# from .date_format import DateFormat
from .list_count import ListCount
from .list_items import ListItems
from .list_mode import ListMode
from .list_order import ListOrder
from .list_search import ListSearch
from .meta_fields import MetaFields
from .meta_filters import MetaFilters
from .meta_orders import MetaOrders
from .subset import Subset
from .subset_limit import SubsetLimit
from .subset_offset import SubsetOffset
from .subset_page import SubsetPage
from .subset_page_size import SubsetPageSize
from .subset_page_subset import SubsetPageSubset


class ModelSerialization(ParameterGroup):
    # TODO: check the need of this params, if needed it should be better
    # to allow to choose format : iso/short/long (for instance).
    # For this reason we choose to not export it in apidoc.
    # parameters = [DateFormat]
    parameters = []


class MetaList(ParameterGroup):
    """Parameters group for meta list services."""
    parameters = [MetaFields, MetaFilters, MetaOrders]


class List(ParameterGroup):
    """Parameters group for list services."""
    parameters = [
        ListCount, ListItems, ListSearch, ListOrder, ListMode,
        Subset, SubsetPage, SubsetPageSize,
        SubsetPageSubset, SubsetOffset, SubsetLimit]



import logging

from django.test import TestCase

from pfx.pfxcore.test import APIClient, TestPermsAssertMixin
from tests.models import Author, User

logger = logging.getLogger(__name__)


class PermTestsTest(TestPermsAssertMixin, TestCase):
    def setUp(self):
        self.client = APIClient(with_cookie=True)

    @classmethod
    def setUpTestData(cls):
        User.objects.create_user(
            username='user',
            email="user@example.com",
            password='test',
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            is_superuser=True)
        User.objects.create_user(
            username='admin',
            email="admin@example.com",
            password='test',
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            is_superuser=True)

    # API calls ---------------------------------------------------------------

    def list(self):
        Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        Author.objects.create(
            first_name='Philip Kindred',
            last_name='Dick',
            slug='philip-k-dick')
        Author.objects.create(
            first_name='Isaac',
            last_name='Asimov',
            slug='isaac-asimov')
        return self.client.get('/api/admin/authors?items=1&count=1')

    def get(self):
        author = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        return self.client.get(f'/api/admin/authors/{author.pk}')

    def post(self):
        return self.client.post('/api/admin/authors', dict(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien'))

    def put(self):
        author = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        return self.client.put(f'/api/admin/authors/{author.pk}', dict(
            first_name='John Ronald Reuel Updated'))

    def delete(self):
        author = Author.objects.create(
            first_name='John Ronald Reuel',
            last_name='Tolkien',
            slug='jrr-tolkien')
        return self.client.delete(f'/api/admin/authors/{author.pk}')

    # Tests -------------------------------------------------------------------

    USER_TESTS = {
        "user": dict(
            list=200, list__count=3,
            get=200,
            post=200,
            put=200,
            delete=200),
        "admin": dict(
            list=200, list__count=3,
            get=200,
            post=200,
            put=200,
            delete=200)}



SECRET_KEY = 'fake-key'
INSTALLED_APPS = [
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.postgres',
    'pfx.pfxcore',
]
DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

import django
from django.conf import settings

from sphinx.ext.autodoc import between

sys.path.insert(0, os.path.abspath('../'))
settings.configure(
    SECRET_KEY='something to make Django happy',
    INSTALLED_APPS=[
        'django.contrib.auth',
        'django.contrib.contenttypes',
        'django.contrib.postgres',
        'pfx.pfxcore'])
django.setup()


def setup(app):
    # Register a sphinx.ext.autodoc.between listener to ignore everything
    # between lines that contain --- and """ (yaml from apispec)
    app.connect('autodoc-process-docstring',
                between('^.*(---|""")$', exclude=True))
    return app


# -- Project information -----------------------------------------------------
project = 'Django PFX'
copyright = '2021, Hervé Martinet, Nicolas Béraneck'
author = 'Hervé Martinet, Nicolas Béraneck'

# The full version, including alpha/beta/rc tags
release = '1.0'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['myst_parser',
              'sphinx.ext.autodoc',
              'sphinx.ext.autosummary',
              'sphinxcontrib.mermaid']
autodoc_member_order = 'bysource'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']

master_doc = 'index'

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

myst_heading_anchors = 3



from datetime import timedelta

from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from pfx.pfxcore.decorator import rest_property
from pfx.pfxcore.settings import settings


class OtpUserMixin(models.Model):
    """A mixin to enable OTP MFA on a user class."""

    #: OTP secret token.
    otp_secret_token = models.CharField(
        _("OTP secret token"), max_length=32, null=True,
        blank=True, unique=True)
    #: Temporary OTP secret token (needs confirmation).
    otp_secret_token_tmp = models.CharField(
        _("Temporary OTP secret token"), max_length=32, null=True, blank=True)
    #: HOTP count.
    hotp_count = models.IntegerField(_("HOTP count"), default=0)
    #: HOTP expiry.
    hotp_expiry = models.DateTimeField(_("HOTP expiry"), default=timezone.now)

    class Meta:
        abstract = True

    @rest_property(_("OTP enabled"), "BooleanField")
    def is_otp(self):
        return bool(self.otp_secret_token)

    def enable_otp(self):
        """Activate OTP for this user.

        Generates a new temporary OTP secret token. To complete activation,
        call `confirm_otp` with a valid code.
        """
        import pyotp
        self.otp_secret_token_tmp = pyotp.random_base32()
        self.save(update_fields=['otp_secret_token_tmp'])

    def confirm_otp(self, otp_code):
        """Confirm OTP activation for this user.

        Set the OTP secret token from the temporary one if the provided
        code is valid.

        :param otp_code: A valid OTP code for the temporary OTP secret key.
        :returns: `True` if success, `False` otherwise.
        """
        if self.is_otp_valid(otp_code, tmp=True):
            self.otp_secret_token = self.otp_secret_token_tmp
            self.otp_secret_token_tmp = None
            self.save(update_fields=[
                'otp_secret_token', 'otp_secret_token_tmp'])
            return True
        return False

    def disable_otp(self):
        """Disable OTP for this user.

        Remove the OTP secret token.
        """
        self.otp_secret_token = None
        self.save(update_fields=['otp_secret_token'])

    def get_otp_setup_uri(self, tmp=False, with_color=True):
        """Return the setup URL for OTP activation.
        """
        import pyotp
        args = dict(
            name=self.get_username(), issuer_name=settings.PFX_SITE_NAME)
        if settings.PFX_OTP_IMAGE:
            args['image'] = settings.PFX_OTP_IMAGE
        uri = pyotp.totp.TOTP(
            tmp and self.otp_secret_token_tmp or
            self.otp_secret_token).provisioning_uri(**args)
        if with_color and settings.PFX_OTP_COLOR:
            # TODO: Can be put in provisioning_uri args if
            # https://github.com/pyauth/pyotp/pull/164 is merge and published.
            uri = f"{uri}&color={settings.PFX_OTP_COLOR}"
        return uri

    def is_otp_valid(self, otp_code, tmp=False):
        """Verify an OTP code.

        :param otp_code: A valid OTP code for the OTP secret key.
        :param tmp: If `True`, verify the code with the temporary
            OTP secret key.
        :returns: `True` if the code is valid, `False` otherwise.
        """
        import pyotp

        # TODO : The with_color paramter can be removed if
        # https://github.com/pyauth/pyotp/pull/164 is merge and published.
        totp = pyotp.parse_uri(
            self.get_otp_setup_uri(tmp=tmp, with_color=False))
        valid = totp.verify(
            otp_code, valid_window=settings.PFX_OTP_VALID_WINDOW)
        if not valid and timezone.now() <= self.hotp_expiry:
            hotp = pyotp.hotp.HOTP(
                tmp and self.otp_secret_token_tmp or
                self.otp_secret_token)
            return hotp.verify(otp_code, self.hotp_count)
        return valid

    def get_user_jwt_signature_key(self):
        """Return a user secret to sign JWT token.

        If the user inherit :class:`pfx.pfxcore.models.AbstractPFXBaseUser`,
        add the OTP secret token to the user signature."""
        return super().get_user_jwt_signature_key() + (
            self.otp_secret_token or "")

    def get_hotp_code(self):
        """Return a new valid HOTP code.

        Increment the HOTP counter and reset the expiry."""
        import pyotp
        if not self.otp_secret_token:
            raise Exception("OTP disabled")
        self.hotp_count += 1
        self.hotp_expiry = timezone.now() + timedelta(
            minutes=settings.PFX_HOTP_CODE_VALIDITY)
        self.save(update_fields=[
            'hotp_count', 'hotp_expiry'])
        return pyotp.hotp.HOTP(self.otp_secret_token).at(self.hotp_count)



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class SubsetPage(QueryParameter):
    """The page number to query in `pagination` mode.

    If the value of `subset` is not `"pagination"`, this attribute will
    be ignored.
    """
    name = 'page'
    schema = dict(type='integer', default=1)


register_global_parameter(SubsetPage)



import logging
from unittest.mock import patch

from django.conf import settings
from django.test import TestCase
from django.test.utils import override_settings

from pfx.pfxcore.middleware import profiling
from pfx.pfxcore.test import APIClient, TestAssertMixin

logger = logging.getLogger(__name__)


@override_settings(
    DEBUG=True,
    MIDDLEWARE=[
        'pfx.pfxcore.middleware.ProfilingMiddleware'
    ] + settings.MIDDLEWARE)
class ProfilingMiddlewareTest(TestAssertMixin, TestCase):
    def setUp(self):
        self.client = APIClient()

    @patch.object(profiling.logger, 'debug')
    def test_profile(self, mock_debug):
        response = self.client.get('/api/authors?items=1&count=1')
        self.assertRC(response, 200)
        self.assertTrue(mock_debug.called)
        log = mock_debug.call_args[0][0]
        self.assertIn('GET /api/authors?items=1&count=1 200', log)
        self.assertIn('SQL queries: 2', log)
        self.assertNotIn('SELECT COUNT(*)', log)
        self.assertNotIn('SELECT DISTINCT', log)
        self.assertNotIn('Ordered by: cumulative time', log)
        self.assertNotIn('ncalls  tottime  percall  cumtime  percall', log)

    @override_settings(
        PFX_PROFILING_SQL='/api/authors')
    @patch.object(profiling.logger, 'debug')
    def test_profile_sql(self, mock_debug):
        response = self.client.get('/api/authors?items=1&count=1')
        self.assertRC(response, 200)
        self.assertTrue(mock_debug.called)
        log = mock_debug.call_args[0][0]
        self.assertIn('GET /api/authors?items=1&count=1 200', log)
        self.assertIn('SQL queries: 2', log)
        self.assertIn('SELECT COUNT(*)', log)
        self.assertIn('SELECT DISTINCT', log)

    @override_settings(
        PFX_PROFILING_SQL='/api/books')
    @patch.object(profiling.logger, 'debug')
    def test_profile_sql_other_path(self, mock_debug):
        response = self.client.get('/api/authors?items=1&count=1')
        self.assertRC(response, 200)
        self.assertTrue(mock_debug.called)
        log = mock_debug.call_args[0][0]
        self.assertIn('GET /api/authors?items=1&count=1 200', log)
        self.assertIn('SQL queries: 2', log)
        self.assertNotIn('SELECT COUNT(*)', log)
        self.assertNotIn('SELECT DISTINCT', log)

    @override_settings(
        PFX_PROFILING_CPROFILE='/api/authors')
    @patch.object(profiling.logger, 'debug')
    def test_profile_cprofile(self, mock_debug):
        response = self.client.get('/api/authors?items=1&count=1')
        self.assertRC(response, 200)
        self.assertTrue(mock_debug.called)
        log = mock_debug.call_args[0][0]
        # print(log)
        self.assertIn('GET /api/authors?items=1&count=1 200', log)
        self.assertIn('SQL queries: 2', log)
        self.assertIn('Ordered by: cumulative time', log)
        self.assertIn('ncalls  tottime  percall  cumtime  percall', log)

    @override_settings(
        PFX_PROFILING_CPROFILE='/api/books')
    @patch.object(profiling.logger, 'debug')
    def test_profile_cprofile_other_path(self, mock_debug):
        response = self.client.get('/api/authors?items=1&count=1')
        self.assertRC(response, 200)
        self.assertTrue(mock_debug.called)
        log = mock_debug.call_args[0][0]
        self.assertIn('GET /api/authors?items=1&count=1 200', log)
        self.assertIn('SQL queries: 2', log)
        self.assertNotIn('Ordered by: cumulative time', log)
        self.assertNotIn('ncalls  tottime  percall  cumtime  percall', log)



import logging

from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.core.serializers.json import DjangoJSONEncoder

logger = logging.getLogger(__name__)


class PFXJSONEncoder(DjangoJSONEncoder):
    def default(self, obj):
        if isinstance(obj, ValidationError):
            if hasattr(obj, 'error_dict'):
                return dict(obj)
            return {NON_FIELD_ERRORS: list(obj)}
        return super().default(obj)



import logging
import re
from datetime import datetime, timedelta
from http.cookies import SimpleCookie
from urllib.parse import parse_qs, urlparse

from django.conf import settings
from django.contrib.auth.tokens import default_token_generator
from django.core import mail
from django.test import TransactionTestCase, modify_settings, override_settings
from django.utils import timezone
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode

import jwt
import pyotp
from freezegun import freeze_time

from pfx.pfxcore.test import APIClient, TestAssertMixin, get_auth_cookie
from tests.models import User

logger = logging.getLogger(__name__)


class AuthAPITest(TestAssertMixin, TransactionTestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')
        self.cookie_client = APIClient(default_locale='en', with_cookie=True)
        self.user1 = User.objects.create_user(
            username='jrr.tolkien',
            email="jrr.tolkien@oxford.com",
            password='RIGHT PASSWORD',
            first_name='John Ronald Reuel',
            last_name='Tolkien',
        )

    def test_invalid_login(self):
        response = self.client.post(
            '/api/auth/login', {
                'username': 'jrr.tolkien',
                'password': 'WRONG PASSWORD'})
        self.assertRC(response, 422)

    def test_emtpy_login(self):
        response = self.client.post(
            '/api/auth/login', {})
        self.assertRC(response, 422)

    @override_settings(PFX_TOKEN_SHORT_VALIDITY={'minutes': 30})
    def test_valid_login(self):
        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post(
                '/api/auth/login', {
                    'username': 'jrr.tolkien',
                    'password': 'RIGHT PASSWORD'})

            self.assertRC(response, 200)
            token = self.get_val(response, 'token')
            headers = jwt.get_unverified_header(token)
            self.assertEqual(headers['pfx_user_pk'], self.user1.pk)
            decoded = jwt.decode(
                token, self.user1.password + settings.PFX_SECRET_KEY,
                algorithms="HS256")
            self.assertTrue(
                datetime.utcnow() + timedelta(minutes=25) <
                datetime.utcfromtimestamp(decoded['exp'])
                < datetime.utcnow() + timedelta(minutes=35))

    @override_settings(PFX_TOKEN_LONG_VALIDITY={'minutes': 60})
    def test_valid_login_remember_me(self):
        response = self.client.post(
            '/api/auth/login', {
                'username': 'jrr.tolkien',
                'password': 'RIGHT PASSWORD',
                'remember_me': True})

        self.assertRC(response, 200)
        token = self.get_val(response, 'token')
        headers = jwt.get_unverified_header(token)
        self.assertEqual(headers['pfx_user_pk'], self.user1.pk)
        decoded = jwt.decode(
            token, self.user1.password + settings.PFX_SECRET_KEY,
            algorithms="HS256")
        self.assertTrue(
            datetime.utcnow() + timedelta(minutes=55) <
            datetime.utcfromtimestamp(decoded['exp'])
            < datetime.utcnow() + timedelta(minutes=65))

    @override_settings(
        PFX_TOKEN_SHORT_VALIDITY={'minutes': 30},
        PFX_LOGIN_BAN_FAILED_NUMBER=2,
        PFX_LOGIN_BAN_SECONDS_START=30,
        PFX_LOGIN_BAN_SECONDS_STEP=60)
    def test_login_ban(self):
        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post(
                '/api/auth/login', {
                    'username': 'jrr.tolkien',
                    'password': 'BAD PASSWORD'})
            self.assertRC(response, 422)
            response = self.client.post(
                '/api/auth/login', {
                    'username': 'jrr.tolkien',
                    'password': 'BAD PASSWORD'})
            self.assertRC(response, 422)

            response = self.client.post(
                '/api/auth/login', {
                    'username': 'jrr.tolkien',
                    'password': 'BAD PASSWORD'})
            self.assertRC(response, 429)
            self.assertEqual(response.headers['Retry-After'], '30')
            self.assertJE(
                response, 'message',
                "Your connection is temporarily disabled after several "
                "unsuccessful attempts, please retry in 30 seconds.")

        with freeze_time("2023-05-01 08:00:20"):
            response = self.client.post(
                '/api/auth/login', {
                    'username': 'jrr.tolkien',
                    'password': 'BAD PASSWORD'})
            self.assertRC(response, 429)
            self.assertEqual(response.headers['Retry-After'], '10')
            self.assertJE(
                response, 'message',
                "Your connection is temporarily disabled after several "
                "unsuccessful attempts, please retry in 10 seconds.")

        with freeze_time("2023-05-01 08:00:30"):
            response = self.client.post(
                '/api/auth/login', {
                    'username': 'jrr.tolkien',
                    'password': 'BAD PASSWORD'})
            self.assertRC(response, 422)
            response = self.client.post(
                '/api/auth/login', {
                    'username': 'jrr.tolkien',
                    'password': 'BAD PASSWORD'})
            self.assertRC(response, 422)

            response = self.client.post(
                '/api/auth/login', {
                    'username': 'jrr.tolkien',
                    'password': 'BAD PASSWORD'})
            self.assertRC(response, 429)
            self.assertEqual(response.headers['Retry-After'], '90')
            self.assertJE(
                response, 'message',
                "Your connection is temporarily disabled after several "
                "unsuccessful attempts, please retry in 90 seconds.")

    @override_settings(PFX_TOKEN_SHORT_VALIDITY={'minutes': 30})
    def test_token_invalid_when_password_change(self):
        response = self.client.post(
            '/api/auth/login', {
                'username': 'jrr.tolkien',
                'password': 'RIGHT PASSWORD'})
        self.assertRC(response, 200)
        token = self.get_val(response, 'token')

        # Token is valid
        response = self.client.get(
            '/api/private/authors',
            HTTP_AUTHORIZATION='Bearer ' + token,
            content_type='application/json')
        self.assertRC(response, 200)

        # Token is invalid after password change
        self.user1.password = "NEW PASSWORD"
        self.user1.save()
        response = self.client.get(
            '/api/private/authors',
            HTTP_AUTHORIZATION='Bearer ' + token,
            content_type='application/json')
        self.assertRC(response, 401)

    @override_settings(PFX_TOKEN_SHORT_VALIDITY={'minutes': 30})
    def test_valid_login_with_cookie(self):
        response = self.client.post(
            '/api/auth/login?mode=cookie', {
                'username': 'jrr.tolkien',
                'password': 'RIGHT PASSWORD'})

        cookie = get_auth_cookie(response)
        regex = r".*token=([\w\._-]*);.*"
        token = re.findall(regex, str(cookie))[0]

        # The cookie should be a session cookie
        # (removed when browser is closed)
        regex = r".*expires=([^;]*);.*"
        expires = re.findall(regex, str(cookie))
        regex = r".*Max-Age=([^;]*);.*"
        max_age = re.findall(regex, str(cookie))
        self.assertEqual(len(expires), 0)
        self.assertEqual(len(max_age), 0)

        self.assertRC(response, 200)
        headers = jwt.get_unverified_header(token)
        self.assertEqual(headers['pfx_user_pk'], self.user1.pk)
        decoded = jwt.decode(
            token, self.user1.password + settings.PFX_SECRET_KEY,
            algorithms="HS256")

        # token expires in 30 minutes +/- 5 minutes.
        self.assertTrue(
            datetime.utcnow() + timedelta(minutes=25) <
            datetime.utcfromtimestamp(decoded['exp'])
            < datetime.utcnow() + timedelta(minutes=35))

    @override_settings(PFX_TOKEN_LONG_VALIDITY={'minutes': 60})
    def test_valid_login_with_cookie_remember_me(self):
        response = self.client.post(
            '/api/auth/login?mode=cookie', {
                'username': 'jrr.tolkien',
                'password': 'RIGHT PASSWORD',
                'remember_me': 'true'})

        cookie = get_auth_cookie(response)

        regex = r".*token=([\w\._-]*);.*"
        token = re.findall(regex, str(cookie))[0]

        regex = r".*expires=([^;]*);.*"
        expires = re.findall(regex, str(cookie))[0]
        d = datetime.strptime(expires, '%a, %d %b %Y %H:%M:%S %Z')

        regex = r".*Max-Age=([^;]*);.*"
        max_age = re.findall(regex, str(cookie))[0]

        # cookie expires in 60 minutes +/- 5 minutes.
        self.assertTrue(
            datetime.utcnow() + timedelta(minutes=55) <
            d < datetime.utcnow() + timedelta(minutes=65))
        self.assertEqual(int(max_age), 3600)

        self.assertRC(response, 200)
        headers = jwt.get_unverified_header(token)
        self.assertEqual(headers['pfx_user_pk'], self.user1.pk)
        decoded = jwt.decode(
            token, self.user1.password + settings.PFX_SECRET_KEY,
            algorithms="HS256")

        # token expires in 60 minutes +/- 5 minutes.
        self.assertTrue(
            datetime.utcnow() + timedelta(minutes=55) <
            datetime.utcfromtimestamp(decoded['exp'])
            < datetime.utcnow() + timedelta(minutes=65))

    def test_logout(self):
        self.client.login(
            username='jrr.tolkien',
            password='RIGHT PASSWORD')
        response = self.client.get('/api/auth/logout')
        self.assertRC(response, 200)
        cookie = [v for k, v in response.client.cookies.items()
                  if k == 'token'][0]
        regex = r".*token=([\"\"]*);.*"
        token = re.findall(regex, str(cookie))[0]
        self.assertEqual(token, '""')

    @override_settings(
        PFX_COOKIE_DOMAIN='example.com',
        PFX_COOKIE_SECURE='True',
        PFX_COOKIE_SAMESITE='Lax'
    )
    def test_logout_with_cookie_params(self):
        self.client.login(
            username='jrr.tolkien',
            password='RIGHT PASSWORD')
        response = self.client.get('/api/auth/logout')
        self.assertRC(response, 200)
        cookie = [v for k, v in response.client.cookies.items()
                  if k == 'token'][0]

        regex = r"token=([\"\"]*);"
        token = re.findall(regex, str(cookie))[0]
        self.assertEqual(token, '""')

        regex = r"Domain=(.*?);"
        domain = re.findall(regex, str(cookie))[0]
        self.assertEqual(domain, 'example.com')

        regex = r"SameSite=(.*)"
        sameSite = re.findall(regex, str(cookie))[0]
        self.assertEqual(sameSite, 'Lax')

    def test_valid_change_password(self):
        self.client.login(
            username='jrr.tolkien',
            password='RIGHT PASSWORD')
        response = self.client.post(
            '/api/auth/change-password', {
                'old_password': 'RIGHT PASSWORD',
                'new_password': 'NEW RIGHT PASSWORD'})
        self.assertRC(response, 200)

        self.cookie_client.login(
            username='jrr.tolkien',
            password='NEW RIGHT PASSWORD')
        response = self.cookie_client.post(
            '/api/auth/change-password', {
                'old_password': 'NEW RIGHT PASSWORD',
                'new_password': 'RIGHT PASSWORD'})
        self.assertRC(response, 200)

    @override_settings(AUTH_PASSWORD_VALIDATORS=[{
        'NAME':
            'django.contrib.auth.password_validation.'
            'UserAttributeSimilarityValidator',
    }, {
        'NAME': 'django.contrib.auth.password_validation.'
                'MinimumLengthValidator',
    }])
    def test_change_password_validation(self):
        self.client.login(
            username='jrr.tolkien',
            password='RIGHT PASSWORD')
        response = self.client.post(
            '/api/auth/change-password', {
                'old_password': 'RIGHT PASSWORD',
                'new_password': 'jrr'})
        self.assertRC(response, 422)
        self.assertJE(response, 'new_password.@0',
                      "The password is too similar to the username.")
        self.assertJE(response, 'new_password.@1',
                      "This password is too short. "
                      "It must contain at least 8 characters.")
        response = self.client.post(
            '/api/auth/change-password', {
                'old_password': 'RIGHT PASSWORD',
                'new_password': '9ashff8za-@#asd'})
        self.assertRC(response, 200)

    def test_invalid_change_password(self):
        self.client.login(
            username='jrr.tolkien',
            password='RIGHT PASSWORD')

        # Wrong old password
        response = self.client.post(
            '/api/auth/change-password', {
                'old_password': 'WRONG PASSWORD',
                'new_password': 'NEVER APPLIED PASSWORD'})
        self.assertRC(response, 422)
        self.assertJE(response, "old_password.@0", "Incorrect password")

    def test_empty_change_password(self):
        self.client.login(
            username='jrr.tolkien',
            password='RIGHT PASSWORD')
        response = self.client.post(
            '/api/auth/change-password', {})
        self.assertRC(response, 422)
        self.assertJE(response, "old_password.@0", "Incorrect password")
        self.assertJE(
            response, "new_password.@0", "Empty password is not allowed")

        # No new password
        response = self.client.post(
            '/api/auth/change-password', {
                'old_password': 'RIGHT PASSWORD'})
        self.assertRC(response, 422)
        self.assertJE(
            response, "new_password.@0", "Empty password is not allowed")

        # Empty new password
        response = self.client.post(
            '/api/auth/change-password', {
                'old_password': 'RIGHT PASSWORD',
                'new_password': ''})
        self.assertRC(response, 422)
        self.assertJE(
            response, "new_password.@0", "Empty password is not allowed")

        # None new password
        response = self.client.post(
            '/api/auth/change-password', {
                'old_password': 'RIGHT PASSWORD',
                'new_password': None})
        self.assertRC(response, 422)
        self.assertJE(
            response, "new_password.@0", "Empty password is not allowed")

    @override_settings(
        PFX_TOKEN_SHORT_VALIDITY={'minutes': 30},
        PFX_COOKIE_DOMAIN='example.com')
    def test_valid_cookie_token_with_domain(self):
        self.cookie_client.login(
            username='jrr.tolkien',
            password='RIGHT PASSWORD')

        regex = r".*Domain=([^;]*);.*"
        domain = re.findall(regex, str(self.cookie_client.auth_cookie))[0]
        self.assertEqual(domain, 'example.com')

        regex = r"\s(Secure)\s*"
        secure = re.findall(regex, str(self.cookie_client.auth_cookie))[0]
        self.assertEqual(secure, "Secure")  # default value

        regex = r".*SameSite=([^;]*).*"
        samesite = re.findall(regex, str(self.cookie_client.auth_cookie))[0]
        self.assertEqual(samesite, "None")  # default value

        response = self.cookie_client.get(
            '/api/private/authors')
        self.assertRC(response, 200)

    @override_settings(
        PFX_COOKIE_DOMAIN='example.com',
        PFX_COOKIE_SECURE=False,
        PFX_COOKIE_SAMESITE='Lax')
    def test_valid_cookie_token_with_domain_same_site_insecure(self):
        self.cookie_client.login(
            username='jrr.tolkien',
            password='RIGHT PASSWORD')

        regex = r".*Domain=([^;]*);.*"
        domain = re.findall(regex, str(self.cookie_client.auth_cookie))[0]
        self.assertEqual(domain, 'example.com')

        regex = r"\s(Secure)\s*"
        self.assertEqual(
            len(re.findall(regex, str(self.cookie_client.auth_cookie))), 0)

        regex = r".*SameSite=([^;]*).*"
        samesite = re.findall(regex, str(self.cookie_client.auth_cookie))[0]
        self.assertEqual(samesite, "Lax")

        response = self.cookie_client.get(
            '/api/private/authors')
        self.assertRC(response, 200)

    @override_settings(PFX_TOKEN_SHORT_VALIDITY={'hours': 4})
    def test_expired_token(self):
        with freeze_time("2023-05-01 08:00:00"):
            self.client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD')
        with freeze_time("2023-05-01 12:01:00"):
            response = self.client.get(
                '/api/private/authors')
            self.assertRC(response, 401)

    @override_settings(PFX_TOKEN_LONG_VALIDITY={'days': 2})
    def test_expired_cookie_with_expired_token(self):

        with freeze_time("2023-05-01 08:00:00"):
            self.cookie_client.login(
                username='jrr.tolkien',
                password='RIGHT PASSWORD',
                remember_me=True)
        with freeze_time("2023-05-04 08:00:00"):
            response = self.cookie_client.get(
                '/api/private/authors')
            self.assertRC(response, 401)
            self.assertEqual(
                str(get_auth_cookie(response)),
                'Set-Cookie: token=""; expires=Thu, 01 Jan 1970 00:00:00 GMT; '
                'Max-Age=0; Path=/; SameSite=None; Secure')

    def test_invalid_auth_header(self):
        response = self.client.get(
            '/api/private/authors',
            HTTP_AUTHORIZATION='Beer here',
            content_type='application/json')
        self.assertRC(response, 401)

    def test_valid_token_with_invalid_user(self):
        user = User.objects.create_user(
            username='invisible.man',
            email="iv@invisible.com",
            password='RIGHT PASSWORD',
            first_name='Peter',
            last_name='Invisible',
        )
        self.client.login(
            username='invisible.man',
            password='RIGHT PASSWORD')
        user.delete()
        response = self.client.get(
            '/api/private/authors')
        self.assertRC(response, 401)

    def test_valid_cookie_with_invalid_user(self):
        user = User.objects.create_user(
            username='invisible.man',
            email="iv@invisible.com",
            password='RIGHT PASSWORD',
            first_name='Peter',
            last_name='Invisible',
        )
        self.cookie_client.login(
            username='invisible.man',
            password='RIGHT PASSWORD')
        user.delete()
        response = self.cookie_client.get(
            '/api/private/authors')
        self.assertRC(response, 401)
        self.assertEqual(
            str(get_auth_cookie(response)),
            'Set-Cookie: token=""; expires=Thu, 01 Jan 1970 00:00:00 GMT; '
            'Max-Age=0; Path=/; SameSite=None; Secure')

    def test_valid_cookie_with_invalid_user_and_public_service(self):
        user = User.objects.create_user(
            username='invisible.man',
            email="iv@invisible.com",
            password='RIGHT PASSWORD',
            first_name='Peter',
            last_name='Invisible',
        )
        self.cookie_client.login(
            username='invisible.man',
            password='RIGHT PASSWORD')
        user.delete()
        response = self.cookie_client.get(
            '/api/books')
        self.assertRC(response, 200)
        self.assertEqual(
            str(get_auth_cookie(response)),
            'Set-Cookie: token=""; expires=Thu, 01 Jan 1970 00:00:00 GMT; '
            'Max-Age=0; Path=/; SameSite=None; Secure')

    def test_invalid_token(self):
        token = jwt.encode(
            {'pfx_user_pk': 1}, "A WRONG SECRET", algorithm="HS256")
        response = self.client.get(
            '/api/private/authors',
            HTTP_AUTHORIZATION='Bearer ' + token,
            content_type='application/json')
        self.assertRC(response, 401)

    def test_invalid_token_with_public_service(self):
        token = jwt.encode(
            {'pfx_user_pk': 1}, "A WRONG SECRET", algorithm="HS256")
        response = self.client.get(
            '/api/books',
            HTTP_AUTHORIZATION='Bearer ' + token,
            content_type='application/json')
        self.assertRC(response, 200)

    def test_invalid_cookie_token(self):
        token = jwt.encode(
            {'pfx_user_pk': 1}, "A WRONG SECRET", algorithm="HS256")
        self.client.cookies = SimpleCookie({'token': token})
        response = self.client.get(
            '/api/private/authors',
            content_type='application/json')
        self.assertRC(response, 401)
        self.assertEqual(
            str(get_auth_cookie(response)),
            'Set-Cookie: token=""; expires=Thu, 01 Jan 1970 00:00:00 GMT; '
            'Max-Age=0; Path=/; SameSite=None; Secure')

    def test_invalid_cookie_token_with_public_service(self):
        token = jwt.encode(
            {'pfx_user_pk': 1}, "A WRONG SECRET", algorithm="HS256")
        self.client.cookies = SimpleCookie({'token': token})
        response = self.client.get(
            '/api/books',
            content_type='application/json')
        self.assertRC(response, 200)
        self.assertEqual(
            str(get_auth_cookie(response)),
            'Set-Cookie: token=""; expires=Thu, 01 Jan 1970 00:00:00 GMT; '
            'Max-Age=0; Path=/; SameSite=None; Secure')

    def test_signup(self):
        # Try to create with an existing username
        response = self.client.post(
            '/api/auth/signup', {
                'username': 'jrr.tolkien',
                'email': "jrr.tolkien@oxford.com",
                'first_name': 'John Ronald Reuel',
                'last_name': 'Tolkien',
            })

        self.assertRC(response, 422)
        self.assertEqual(response.json_content['username'],
                         ['A user with that username already exists.'])

        # Then create another valid user
        response = self.client.post(
            '/api/auth/signup', {
                'username': 'isaac.asimov',
                'email': 'isaac.asimov@bu.edu',
                'first_name': 'Isaac',
                'last_name': 'Asimov',
            })

        self.assertRC(response, 200)

        # Must send a welcome email
        self.assertEqual(
            mail.outbox[0].subject,
            f'Welcome on {settings.PFX_SITE_NAME}')

        self.client.logout()

        # Test that the token and uid are valid.
        regex = r"token=(.*)&uidb64=(.*)"
        token, uidb64 = re.findall(regex, mail.outbox[0].body)[0]
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uidb64,
                'password': 'test'
            })
        self.assertRC(response, 200)
        self.assertJE(response, 'message', 'password updated successfully')

    def test_empty_signup(self):
        # Try empty signup
        response = self.client.post(
            '/api/auth/signup', {})
        self.assertRC(response, 422)
        self.assertJE(response, 'username.@0', 'This field cannot be blank.')

    def test_forgotten_password(self):
        # Try with an nonexistent email
        response = self.client.post(
            '/api/auth/forgotten-password', {
                'email': 'isaac.asimov@bu.edu',
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'message',
                      'If the email address you entered is correct, '
                      'you will receive an email from us with '
                      'instructions to reset your password.')

        # Then try with a valid email
        response = self.client.post(
            '/api/auth/forgotten-password', {
                'email': 'jrr.tolkien@oxford.com',
            })

        self.assertRC(response, 200)
        self.assertJE(response, 'message',
                      'If the email address you entered is correct, '
                      'you will receive an email from us with '
                      'instructions to reset your password.')

        # Must send a reset password email
        self.assertEqual(
            mail.outbox[0].subject,
            f'Password reset on {settings.PFX_SITE_NAME}')

        self.client.logout()

        # Test that the token and uid are valid.
        regex = r"token=(.*)&uidb64=(.*)"
        token, uidb64 = re.findall(regex, mail.outbox[0].body)[0]
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uidb64,
                'password': 'test'
            })
        self.assertRC(response, 200)
        self.assertJE(response, 'message', 'password updated successfully')

    def test_empty_email_forgotten_password(self):
        # Try with an empty body
        response = self.client.post(
            '/api/auth/forgotten-password', {
            })
        self.assertRC(response, 422)
        self.assertJE(response, 'email.@0', 'Enter a valid email address.')

        # Try with an empty email address
        response = self.client.post(
            '/api/auth/forgotten-password', {
                'email': '',
            })
        self.assertRC(response, 422)
        self.assertJE(response, 'email.@0', 'Enter a valid email address.')

    def test_wrong_email_forgotten_password(self):
        # Try with an invalid email address
        response = self.client.post(
            '/api/auth/forgotten-password', {
                'email': 'jrr.tolkien',
            })
        self.assertRC(response, 422)
        self.assertJE(response, 'email.@0', 'Enter a valid email address.')

        # Try with another invalid email address
        response = self.client.post(
            '/api/auth/forgotten-password', {
                'email': '@bluestar.solutions',
            })
        self.assertRC(response, 422)
        self.assertJE(response, 'email.@0', 'Enter a valid email address.')

    def test_validate_user_token(self):
        # Try with a wrong token and uid
        response = self.client.post(
            '/api/auth/validate-user-token', {
                'token': 'WRONG TOKEN',
                'uidb64': 'WRONG UID'
            })

        self.assertRC(response, 422)
        self.assertJE(response, '__all__.@0', "User or token is invalid")

        # Then try with a valid token and uid
        token = default_token_generator.make_token(self.user1)
        uid = urlsafe_base64_encode(force_bytes(self.user1.pk))
        response = self.client.post(
            '/api/auth/validate-user-token', {
                'token': token,
                'uidb64': uid
            })
        self.assertRC(response, 200)
        self.assertJE(response, '__all__.@0', "User and token are valid")

    def test_set_password(self):
        # Try with a wrong token and uid
        response = self.client.post(
            '/api/auth/set-password', {
                'token': 'WRONG TOKEN',
                'uidb64': 'WRONG UID',
                'password': 'NEW PASSWORD',
            })
        self.assertRC(response, 401)

        # Then try with a valid token and uid
        token = default_token_generator.make_token(self.user1)
        uid = urlsafe_base64_encode(force_bytes(self.user1.pk))
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uid,
                'password': 'NEW PASSWORD',
            })
        self.assertRC(response, 200)
        self.assertJE(response, 'message', "password updated successfully")

    @override_settings(AUTH_PASSWORD_VALIDATORS=[{
        'NAME':
            'django.contrib.auth.password_validation.'
            'UserAttributeSimilarityValidator',
    }, {
        'NAME': 'django.contrib.auth.password_validation.'
                'MinimumLengthValidator',
    }])
    def test_set_password_validation(self):
        token = default_token_generator.make_token(self.user1)
        uid = urlsafe_base64_encode(force_bytes(self.user1.pk))

        # Missing password
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uid,
            })
        self.assertRC(response, 422)
        self.assertJE(
            response, "password.@0", "Empty password is not allowed")

        # Empty string password
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uid,
                'password': ''
            })
        self.assertRC(response, 422)
        self.assertJE(
            response, "password.@0", "Empty password is not allowed")

        # None password
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uid,
                'password': None
            })
        self.assertRC(response, 422)
        self.assertJE(
            response, "password.@0", "Empty password is not allowed")

        # Invalid password according to UserAttributeSimilarityValidator and
        # MinimumLengthValidator
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uid,
                'password': 'jrr'
            })
        self.assertRC(response, 422)
        self.assertJE(response, 'password.@0',
                      "The password is too similar to the username.")
        self.assertJE(response, 'password.@1',
                      "This password is too short. "
                      "It must contain at least 8 characters.")

        # Finally, a valid password
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uid,
                'password': '9ashff8za-@#asd'
            })
        self.assertRC(response, 200)

    def test_set_password_autologin_jwt(self):
        token = default_token_generator.make_token(self.user1)
        uid = urlsafe_base64_encode(force_bytes(self.user1.pk))
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uid,
                'password': 'NEW PASSWORD',
                'autologin': 'jwt'
            })
        self.assertRC(response, 200)
        self.user1.refresh_from_db()

        self.assertIn('token', response.json_content)
        token = self.get_val(response, 'token')
        headers = jwt.get_unverified_header(token)
        self.assertEqual(headers['pfx_user_pk'], self.user1.pk)
        jwt.decode(
            token, self.user1.password + settings.PFX_SECRET_KEY,
            algorithms="HS256")

    def test_set_password_autologin_cookies(self):
        token = default_token_generator.make_token(self.user1)
        uid = urlsafe_base64_encode(force_bytes(self.user1.pk))
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uid,
                'password': 'NEW PASSWORD',
                'autologin': 'cookie'
            })
        self.assertRC(response, 200)
        self.user1.refresh_from_db()

        cookie = [v for k, v in response.client.cookies.items()
                  if k == 'token'][0]
        regex = r".*token=([\w\._-]*);.*"
        token = re.findall(regex, str(cookie))[0]

        headers = jwt.get_unverified_header(token)
        self.assertEqual(headers['pfx_user_pk'], self.user1.pk)
        jwt.decode(
            token, self.user1.password + settings.PFX_SECRET_KEY,
            algorithms="HS256")

    def test_set_password_autologin_wrong_value(self):
        token = default_token_generator.make_token(self.user1)
        uid = urlsafe_base64_encode(force_bytes(self.user1.pk))
        response = self.client.post(
            '/api/auth/set-password', {
                'token': token,
                'uidb64': uid,
                'password': 'NEW PASSWORD',
                'autologin': 'qwertz'
            })
        self.assertRC(response, 200)
        self.assertJE(response, 'message', "password updated successfully")

    def test_token_anonymous(self):
        self.client.logout()
        response = self.client.get(
            '/api/books')
        self.assertRC(response, 200)

    @modify_settings(
        MIDDLEWARE={
            'remove': [
                'pfx.pfxcore.middleware.AuthenticationMiddleware']})
    def test_cookie_anonymous(self):
        self.cookie_client.logout()
        response = self.cookie_client.get(
            '/api/books')
        self.assertRC(response, 200)

    def enable_otp(self, user):
        user.otp_secret_token = pyotp.random_base32()
        user.save()
        user.refresh_from_db()
        return pyotp.totp.TOTP(self.user1.otp_secret_token)

    def otp_login(self, user, password):
        totp = pyotp.totp.TOTP(user.otp_secret_token)
        self.client.login(username=user.username, password=password)
        response = self.client.post('/api/auth/otp/login', dict(
            token=self.client.token, otp_code=totp.now()))
        self.assertRC(response, 200)
        self.client.token = self.get_val(response, 'token')

    @override_settings(
        PFX_OTP_IMAGE="https://example.org/fake.png",
        PFX_OTP_COLOR="FF0000")
    def test_otp_enable(self):
        self.client.login(username='jrr.tolkien', password='RIGHT PASSWORD')

        self.assertEqual(self.user1.is_otp, False)
        response = self.client.get('/api/auth/otp/setup-uri')
        self.assertRC(response, 200)
        self.assertJIn(response, 'setup_uri', "otpauth://totp/")
        self.assertJIn(response, 'setup_uri', "Books%20Demo:jrr.tolkien")
        self.assertJIn(response, 'setup_uri', "issuer=Books%20Demo")
        self.assertJIn(response, 'setup_uri', "color=FF0000")
        self.assertJIn(
            response, 'setup_uri',
            "image=https%3A%2F%2Fexample.org%2Ffake.png")
        self.user1.refresh_from_db()
        self.assertJIn(
            response, 'setup_uri', f"secret={self.user1.otp_secret_token_tmp}")
        self.assertIsNone(self.user1.otp_secret_token)
        self.assertEqual(len(self.user1.otp_secret_token_tmp), 32)

        url = urlparse(self.get_val(response, 'setup_uri'))
        secret_key = parse_qs(url.query)['secret'][0]
        totp = pyotp.totp.TOTP(secret_key)

        response = self.client.put('/api/auth/otp/enable', dict(
            otp_code=totp.now()))
        self.assertRC(response, 200)
        self.assertJE(response, 'message', "OTP enabled")
        self.user1.refresh_from_db()
        self.assertEqual(len(self.user1.otp_secret_token), 32)
        self.assertIsNone(self.user1.otp_secret_token_tmp)
        self.assertEqual(self.user1.is_otp, True)

    def test_otp_enable_bad_code(self):
        self.client.login(username='jrr.tolkien', password='RIGHT PASSWORD')

        response = self.client.get('/api/auth/otp/setup-uri')
        self.assertRC(response, 200)
        self.assertJIn(response, 'setup_uri', "otpauth://totp/")
        self.user1.refresh_from_db()
        self.assertIsNone(self.user1.otp_secret_token)
        self.assertEqual(len(self.user1.otp_secret_token_tmp), 32)

        response = self.client.put('/api/auth/otp/enable', dict(
            otp_code='-'))
        self.assertRC(response, 422)
        self.assertJE(response, 'otp_code.@0', "Invalid code")
        self.user1.refresh_from_db()
        self.assertIsNone(self.user1.otp_secret_token)
        self.assertEqual(len(self.user1.otp_secret_token_tmp), 32)

    def test_otp_setup_uri_already_activated(self):
        self.enable_otp(self.user1)

        self.otp_login(self.user1, 'RIGHT PASSWORD')
        response = self.client.get('/api/auth/otp/setup-uri')
        self.assertRC(response, 400)
        self.assertJE(response, 'message', "OTP is already enabled")

    def test_otp_disable(self):
        totp = self.enable_otp(self.user1)

        self.assertEqual(self.user1.is_otp, True)
        self.otp_login(self.user1, 'RIGHT PASSWORD')
        response = self.client.put('/api/auth/otp/disable', dict(
            otp_code=totp.now()))
        self.assertRC(response, 200)
        self.assertJE(response, 'message', "OTP disabled")
        self.user1.refresh_from_db()
        self.assertIsNone(self.user1.otp_secret_token)
        self.assertIsNone(self.user1.otp_secret_token_tmp)
        self.assertEqual(self.user1.is_otp, False)

    @override_settings(
        PFX_HOTP_CODE_VALIDITY=15,
        PFX_TOKEN_OTP_VALIDITY={'minutes': 30},)
    def test_otp_disable_by_email(self):
        self.enable_otp(self.user1)

        self.otp_login(self.user1, 'RIGHT PASSWORD')

        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post('/api/auth/otp/email', dict())
            self.assertRC(response, 200)

            self.assertEqual(
                mail.outbox[0].subject,
                f'New authentication code for {settings.PFX_SITE_NAME}')
            code_match = re.search(
                r'Authentication code: (\d{6})', mail.outbox[0].body)
            self.assertIsNotNone(code_match)
            otp_code = code_match.group(1)

            response = self.client.put('/api/auth/otp/disable', dict(
                otp_code=otp_code))
            self.assertRC(response, 200)
            self.assertJE(response, 'message', "OTP disabled")
            self.user1.refresh_from_db()
            self.assertIsNone(self.user1.otp_secret_token)
            self.assertIsNone(self.user1.otp_secret_token_tmp)

    def test_otp_disable_bad_code(self):
        self.enable_otp(self.user1)

        self.otp_login(self.user1, 'RIGHT PASSWORD')
        response = self.client.put('/api/auth/otp/disable', dict(
            otp_code='-'))
        self.assertRC(response, 422)
        self.assertJE(response, 'otp_code.@0', "Invalid code")
        self.user1.refresh_from_db()
        self.assertEqual(len(self.user1.otp_secret_token), 32)
        self.assertIsNone(self.user1.otp_secret_token_tmp)

    @override_settings(PFX_TOKEN_SHORT_VALIDITY={'minutes': 30})
    def test_otp_login(self):
        totp = self.enable_otp(self.user1)

        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post('/api/auth/login', dict(
                username='jrr.tolkien',
                password='RIGHT PASSWORD'))
            self.assertRC(response, 200)
            self.assertJE(response, 'need_otp', True)
            self.assertJEExists(response, 'token')
            token = self.get_val(response, 'token')
            headers = jwt.get_unverified_header(token)
            self.assertEqual(headers['pfx_user_pk'], self.user1.pk)
            decoded = jwt.decode(
                token, self.user1.password + self.user1.otp_secret_token +
                settings.PFX_SECRET_KEY,
                algorithms="HS256")
            # Check the 15m OTP token validity
            self.assertEqual(
                datetime.fromtimestamp(decoded['exp']),
                datetime(2023, 5, 1, 8, 15))

        with freeze_time("2023-05-01 08:10:00"):
            response = self.client.post('/api/auth/otp/login', dict(
                token=token,
                otp_code=totp.now()))
            self.assertRC(response, 200)
            self.assertJE(response, 'user.is_otp', True)
            headers = jwt.get_unverified_header(token)
            self.assertEqual(headers['pfx_user_pk'], self.user1.pk)
            decoded = jwt.decode(
                self.get_val(response, 'token'),
                self.user1.password + self.user1.otp_secret_token +
                settings.PFX_SECRET_KEY,
                algorithms="HS256")
            # Check the short term validity
            self.assertEqual(
                datetime.fromtimestamp(decoded['exp']),
                datetime(2023, 5, 1, 8, 40))

    @override_settings(PFX_TOKEN_LONG_VALIDITY={'days': 1})
    def test_otp_login_remember_me(self):
        totp = self.enable_otp(self.user1)

        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post('/api/auth/login', dict(
                username='jrr.tolkien',
                password='RIGHT PASSWORD',
                remember_me=True))
            self.assertRC(response, 200)
            self.assertJE(response, 'need_otp', True)
            self.assertJEExists(response, 'token')
            token = self.get_val(response, 'token')
            headers = jwt.get_unverified_header(token)
            self.assertEqual(headers['pfx_user_pk'], self.user1.pk)
            decoded = jwt.decode(
                token, self.user1.password + self.user1.otp_secret_token +
                settings.PFX_SECRET_KEY,
                algorithms="HS256")
            # Check the 15m OTP token validity
            self.assertEqual(
                datetime.fromtimestamp(decoded['exp']),
                datetime(2023, 5, 1, 8, 15))

        with freeze_time("2023-05-01 08:10:00"):
            response = self.client.post('/api/auth/otp/login', dict(
                token=token,
                otp_code=totp.now()))
            self.assertRC(response, 200)
            headers = jwt.get_unverified_header(token)
            self.assertEqual(headers['pfx_user_pk'], self.user1.pk)
            decoded = jwt.decode(
                self.get_val(response, 'token'),
                self.user1.password + self.user1.otp_secret_token +
                settings.PFX_SECRET_KEY,
                algorithms="HS256")
            # Check the long term validity
            self.assertEqual(
                datetime.fromtimestamp(decoded['exp']),
                datetime(2023, 5, 2, 8, 10))

    @override_settings(
        PFX_OTP_VALID_WINDOW=0,
        PFX_TOKEN_SHORT_VALIDITY={'minutes': 30})
    def test_otp_login_valid_window_0(self):
        totp = self.enable_otp(self.user1)

        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post('/api/auth/login', dict(
                username='jrr.tolkien',
                password='RIGHT PASSWORD'))
            self.assertRC(response, 200)
            token = self.get_val(response, 'token')

            response = self.client.post('/api/auth/otp/login', dict(
                token=token, otp_code=totp.at(
                    timezone.now() - timedelta(seconds=1))))
            self.assertRC(response, 422)

            response = self.client.post('/api/auth/otp/login', dict(
                token=token, otp_code=totp.at(timezone.now())))
            self.assertRC(response, 200)

    @override_settings(
        PFX_OTP_VALID_WINDOW=1,
        PFX_TOKEN_SHORT_VALIDITY={'minutes': 30})
    def test_otp_login_valid_window_1(self):
        totp = self.enable_otp(self.user1)

        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post('/api/auth/login', dict(
                username='jrr.tolkien',
                password='RIGHT PASSWORD'))
            self.assertRC(response, 200)
            token = self.get_val(response, 'token')

            response = self.client.post('/api/auth/otp/login', dict(
                token=token, otp_code=totp.at(
                    timezone.now() - timedelta(seconds=31))))
            self.assertRC(response, 422)

            response = self.client.post('/api/auth/otp/login', dict(
                token=token, otp_code=totp.at(
                    timezone.now() - timedelta(seconds=30))))
            self.assertRC(response, 200)

    @override_settings(
        PFX_TOKEN_SHORT_VALIDITY={'minutes': 30},
        PFX_LOGIN_BAN_FAILED_NUMBER=2,
        PFX_LOGIN_BAN_SECONDS_START=30,
        PFX_LOGIN_BAN_SECONDS_STEP=60)
    def test_otp_login_ban(self):
        self.enable_otp(self.user1)

        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post('/api/auth/login', dict(
                username='jrr.tolkien',
                password='RIGHT PASSWORD'))
            self.assertRC(response, 200)
            token = self.get_val(response, 'token')

            response = self.client.post('/api/auth/otp/login', dict(
                token=token, otp_code='-'))
            self.assertRC(response, 422)
            response = self.client.post('/api/auth/otp/login', dict(
                token=token, otp_code='-'))
            self.assertRC(response, 422)

            response = self.client.post('/api/auth/otp/login', dict(
                token=token, otp_code='-'))
            self.assertRC(response, 429)
            self.assertEqual(response.headers['Retry-After'], '30')
            self.assertJE(
                response, 'message',
                "Your connection is temporarily disabled after several "
                "unsuccessful attempts, please retry in 30 seconds.")

        with freeze_time("2023-05-01 08:00:20"):
            response = self.client.post('/api/auth/otp/login', dict(
                token=token, otp_code='-'))
            self.assertRC(response, 429)
            self.assertEqual(response.headers['Retry-After'], '10')
            self.assertJE(
                response, 'message',
                "Your connection is temporarily disabled after several "
                "unsuccessful attempts, please retry in 10 seconds.")

        with freeze_time("2023-05-01 08:00:30"):
            response = self.client.post('/api/auth/otp/login', dict(
                token=token, otp_code='-'))
            self.assertRC(response, 422)
            response = self.client.post('/api/auth/otp/login', dict(
                token=token, otp_code='-'))
            self.assertRC(response, 422)

            response = self.client.post('/api/auth/otp/login', dict(
                token=token, otp_code='-'))
            self.assertRC(response, 429)
            self.assertEqual(response.headers['Retry-After'], '90')
            self.assertJE(
                response, 'message',
                "Your connection is temporarily disabled after several "
                "unsuccessful attempts, please retry in 90 seconds.")

    def test_otp_login_expired_token(self):
        totp = self.enable_otp(self.user1)

        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post('/api/auth/login', dict(
                username='jrr.tolkien',
                password='RIGHT PASSWORD'))
            self.assertRC(response, 200)
            self.assertJE(response, 'need_otp', True)
            self.assertJEExists(response, 'token')
            token = self.get_val(response, 'token')

        with freeze_time("2023-05-01 08:16:00"):
            response = self.client.post('/api/auth/otp/login', dict(
                token=token,
                otp_code=totp.now()))
            self.assertRC(response, 401)

    @override_settings(PFX_TOKEN_SHORT_VALIDITY={'minutes': 30})
    def test_otp_login_invalid_code(self):
        self.enable_otp(self.user1)

        response = self.client.post('/api/auth/login', dict(
            username='jrr.tolkien',
            password='RIGHT PASSWORD'))
        self.assertRC(response, 200)
        self.assertJE(response, 'need_otp', True)
        self.assertJEExists(response, 'token')
        token = self.get_val(response, 'token')

        response = self.client.post('/api/auth/otp/login', dict(
            token=token,
            otp_code='-'))
        self.assertRC(response, 422)

    @override_settings(PFX_TOKEN_SHORT_VALIDITY={'minutes': 30})
    def test_otp_login_disabled(self):
        totp = self.enable_otp(self.user1)

        response = self.client.post('/api/auth/login', dict(
            username='jrr.tolkien',
            password='RIGHT PASSWORD'))
        self.assertRC(response, 200)
        self.assertJE(response, 'need_otp', True)
        self.assertJEExists(response, 'token')
        token = self.get_val(response, 'token')

        response = self.client.post('/api/auth/otp/login', dict(
            token=token,
            otp_code=totp.now()))
        self.assertRC(response, 200)
        final_token = self.get_val(response, 'token')

        # Final token is valid
        response = self.client.get(
            '/api/private/authors',
            HTTP_AUTHORIZATION='Bearer ' + final_token,
            content_type='application/json')
        self.assertRC(response, 200)

        self.user1.disable_otp()

        # OTP login is not available
        response = self.client.post('/api/auth/otp/login', dict(
            token=token,
            otp_code=totp.now()))
        self.assertRC(response, 401)

        # Final token is invalid after OTD disabled
        self.user1.password = "NEW PASSWORD"
        self.user1.save()
        response = self.client.get(
            '/api/private/authors',
            HTTP_AUTHORIZATION='Bearer ' + token,
            content_type='application/json')
        self.assertRC(response, 401)

    def test_otp_token_rejected(self):
        totp = self.enable_otp(self.user1)

        response = self.client.post('/api/auth/login', dict(
            username='jrr.tolkien',
            password='RIGHT PASSWORD'))
        self.assertRC(response, 200)
        self.assertJE(response, 'need_otp', True)
        self.assertJEExists(response, 'token')
        otp_token = self.get_val(response, 'token')

        # OTP token is rejected by middleware for other services
        response = self.client.get(
            '/api/private/authors',
            HTTP_AUTHORIZATION='Bearer ' + otp_token)
        self.assertRC(response, 401)

        response = self.client.post('/api/auth/otp/login', dict(
            token=otp_token,
            otp_code=totp.now()))
        self.assertRC(response, 200)
        self.assertJEExists(response, 'token')
        login_token = self.get_val(response, 'token')

        # Login token is accepted
        response = self.client.get(
            '/api/private/authors',
            HTTP_AUTHORIZATION='Bearer ' + login_token)
        self.assertRC(response, 200)

    @override_settings(PFX_HOTP_CODE_VALIDITY=15)
    def test_send_otp_email(self):
        self.enable_otp(self.user1)

        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post('/api/auth/login', dict(
                username='jrr.tolkien',
                password='RIGHT PASSWORD'))
            self.assertRC(response, 200)
            self.assertJE(response, 'need_otp', True)
            self.assertJEExists(response, 'token')
            otp_token = self.get_val(response, 'token')

            response = self.client.post('/api/auth/otp/email', dict(
                token=otp_token))
            self.assertRC(response, 200)

            self.assertEqual(
                mail.outbox[0].subject,
                f'New authentication code for {settings.PFX_SITE_NAME}')
            code_match = re.search(
                r'Authentication code: (\d{6})', mail.outbox[0].body)
            self.assertIsNotNone(code_match)
            otp_code = code_match.group(1)

            response = self.client.post('/api/auth/otp/login', dict(
                token=otp_token,
                otp_code=otp_code))
            self.assertRC(response, 200)
            self.assertJEExists(response, 'token')
            login_token = self.get_val(response, 'token')

            # Login token is accepted
            response = self.client.get(
                '/api/private/authors',
                HTTP_AUTHORIZATION='Bearer ' + login_token)
            self.assertRC(response, 200)

    @override_settings(
        PFX_HOTP_CODE_VALIDITY=15,
        PFX_TOKEN_OTP_VALIDITY={'minutes': 30},)
    def test_send_otp_email_expiry(self):
        self.enable_otp(self.user1)

        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post('/api/auth/login', dict(
                username='jrr.tolkien',
                password='RIGHT PASSWORD'))
            self.assertRC(response, 200)
            self.assertJE(response, 'need_otp', True)
            self.assertJEExists(response, 'token')
            otp_token = self.get_val(response, 'token')

            response = self.client.post('/api/auth/otp/email', dict(
                token=otp_token))
            self.assertRC(response, 200)

            self.assertEqual(
                mail.outbox[0].subject,
                f'New authentication code for {settings.PFX_SITE_NAME}')
            code_match = re.search(
                r'Authentication code: (\d{6})', mail.outbox[0].body)
            self.assertIsNotNone(code_match)
            otp_code = code_match.group(1)

        with freeze_time("2023-05-01 08:15:00"):
            response = self.client.post('/api/auth/otp/login', dict(
                token=otp_token,
                otp_code=otp_code))
            self.assertRC(response, 200)
            self.assertJEExists(response, 'token')

        with freeze_time("2023-05-01 08:15:01"):
            response = self.client.post('/api/auth/otp/login', dict(
                token=otp_token,
                otp_code=otp_code))
            self.assertRC(response, 422)

    @override_settings(
        PFX_HOTP_CODE_VALIDITY=15,
        PFX_TOKEN_OTP_VALIDITY={'minutes': 30},)
    def test_send_otp_email_without_token(self):
        self.enable_otp(self.user1)

        with freeze_time("2023-05-01 08:00:00"):
            response = self.client.post('/api/auth/otp/email', dict())
            self.assertRC(response, 401)



from pfx.pfxcore.apidoc import QueryParameter, register_global_parameter


class ListOrder(QueryParameter):
    """Control the order of objects in result.

    A comma separated list of field, optionally prefixed by :
    * `+` (default) : in ascending order.
    * `-` : in descending order.

    Exemple: `order=-date,name` (`order=-date,+name` is equivalent).

    The list of usable fields can be found in `/meta/list` service response.
    """
    name = 'order'
    schema = dict(type='string', default=None)


register_global_parameter(ListOrder)



from distutils.command.sdist import sdist as sdist_orig

import setuptools


class sdist(sdist_orig):
    def run(self):
        self.spawn(['ls', 'lopp'])
        # super().run()


setuptools.setup(
    use_scm_version=dict(local_scheme="no-local-version"),
    setup_requires=['setuptools_scm']
)  # options are in setup.cfg



import logging

from django.core.exceptions import ImproperlyConfigured
from django.db.models import Count, Q, Value
from django.db.models.functions import Coalesce
from django.utils.translation import gettext as _

from pfx.pfxcore.decorator import rest_api, rest_doc, rest_view
from pfx.pfxcore.http import JsonResponse
from pfx.pfxcore.views import (
    VF,
    BaseRestView,
    FieldType,
    Filter,
    FilterGroup,
    MediaPermsRestViewMixin,
    MediaRestViewMixin,
    ModelFilter,
    PermsRestView,
    RestView,
    SlugDetailRestViewMixin,
    SlugPermsDetailRestViewMixin,
)

from .models import Author, Book, BookType

logger = logging.getLogger(__name__)


class AuthorRestViewMixin():
    model = Author
    fields = [
        'first_name', 'last_name', 'name_length', 'slug', 'gender', 'types',
        VF('books', readonly=True), VF('created_at', readonly=True),
        VF('create_comment', readonly_update=True),
        VF('update_comment', readonly_create=True), 'website']
    list_fields = ['first_name', 'last_name', 'gender']


def heroic_fantasy_filter(value):
    q = Q(last_name="Tolkien")
    return value and q or ~q


def last_name_filter(value):
    return Q(last_name__isnull=not value)


def last_name_choices_filter(value):
    return Q(last_name=value)


def pub_date_gte_filter(value):
    return Q(pub_date__gte=value)


def author_pk_filter(value):
    return Q(author__pk=value)


@rest_view("/authors")
class AuthorRestView(AuthorRestViewMixin, SlugDetailRestViewMixin, RestView):
    default_public = True
    filters = [
        FilterGroup('book_type', _("Book Type"), [
            ModelFilter(Author, 'science_fiction', technical=True),
            Filter(
                'heroic_fantasy', _("Heroic Fantasy"),
                FieldType.BooleanField, heroic_fantasy_filter),
            ModelFilter(Author, 'types', related_model_api='/book-types')
        ]),
        FilterGroup('custom', _("Custom"), [
            ModelFilter(
                Author, 'last_name', type=FieldType.BooleanField,
                filter_func=last_name_filter),
            ModelFilter(Author, 'first_name'),
            ModelFilter(Author, 'gender'),
            Filter(
                'last_name_choices', _("Tolkien or Asimov"),
                FieldType.CharField, last_name_choices_filter,
                choices=[('Tolkien', "Tolkien"), ('Asimov', "Asimov")],
                empty_value=False),
        ]),
    ]

    @rest_api("/cache/<int:id>", method="get", groups=['cache'])
    def cache_get(self, id, *args, **kwargs):
        book = Author.cache_get(id)
        if book:
            return self.response(book, from_cache=True)
        book = self.get_object(pk=id)
        book.cache()
        return self.response(book, from_cache=False)

    @rest_api("/priority/default", method="get")
    def priority_default(self, *args, **kwargs):
        return JsonResponse(dict(value='static:default'))

    @rest_api("/priority/<value>", method="get")
    def priority_param(self, value, *args, **kwargs):
        return JsonResponse(dict(value=f"untyped:{value}"))

    @rest_api("/priority/<int:value>", method="get")
    def priority_int(self, value, *args, **kwargs):
        return JsonResponse(dict(value=f"int:{value}"))

    @rest_api("/priority/<str:value>", method="get")
    def priority_str(self, value, *args, **kwargs):
        return JsonResponse(dict(value=f"str:{value}"))

    @rest_api("/priority/<path:value>", method="get")
    def priority_path(self, value, *args, **kwargs):
        return JsonResponse(dict(value=f"path:{value}"))

    @rest_api("/priority/default/path", method="get")
    def priority_default_path(self, *args, **kwargs):
        return JsonResponse(dict(value="static:default_path"))

    @rest_api("/priority/priority-less", method="get", priority=-1)
    def priority_test2(self, *args, **kwargs):
        return JsonResponse(dict(value='static:less'))

    @rest_api("/priority/priority-more", method="get", priority=1)
    def priority_test3(self, *args, **kwargs):
        return JsonResponse(dict(value='static:more'))


@rest_view("/authors-extra-meta")
class AuthorExtraMetaRestView(AuthorRestViewMixin, RestView):
    default_public = True

    def _get_list_extra_meta(self):
        return dict(test="Hello world")


def books_json_repr(book):
    """JSON representation.
    ---
    _extends: true
    author_gender:
        type: string
    """
    return book.json_repr(author_gender=book.author.gender)


@rest_doc('/<int:id>', 'get', summary="Get custom author", groups=["custom"])
@rest_view("/authors-annotate")
class AuthorAnnotateRestView(AuthorRestView):
    fields = [
        'first_name', 'last_name', 'slug',
        VF('books', json_repr=books_json_repr),
        'books_count', 'books_count_annotate', 'books_count_prop']
    list_fields = [
        'first_name', 'last_name', 'slug',
        'books_count', 'books_count_annotate', 'books_count_prop']

    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).annotate(
            books_count_annotate=Count('books'))


@rest_view("/authors-fields-props")
class AuthorFieldsPropsRestView(AuthorRestView):
    fields = [
        'first_name', 'last_name', 'slug',
        VF('books_count', "Books Count", "IntegerField"),
        VF('books_count_annotate', "Books Count (annotate)", "IntegerField"),
        VF('books_count_prop', "Books Count (property)", "IntegerField")]
    list_fields = [
        'first_name', 'last_name', 'slug',
        VF('books_count', "B.C.", "IntegerField"),
        VF('books_count_annotate', "B.C. (a)", "IntegerField"),
        VF('books_count_prop', "B.C. (p)", "IntegerField")]

    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).annotate(
            books_count_annotate=Count('books'))


@rest_view("/private-edit/authors")
class PrivateEditAuthorRestView(AuthorRestViewMixin, RestView):
    get_public = True
    get_list_public = True


@rest_view("/private/authors")
class PrivateAuthorRestView(AuthorRestViewMixin, RestView):
    pass


@rest_view("/perms/authors")
class PermsAuthorRestView(
        AuthorRestViewMixin, SlugPermsDetailRestViewMixin, PermsRestView):

    @rest_api("/custom", method="get", perms='tests.view_author')
    def get_custom(self):
        return JsonResponse(dict(message='Custom get'))

    @rest_api("/custom", method="put", perms=[
        'tests.change_author', 'tests.can_customize_author'])
    def get_custom_action(self):
        return JsonResponse(dict(message='Custom put'))


@rest_view("/admin-edit/authors")
class AdminEditAuthorRestView(AuthorRestViewMixin, RestView):
    def put_perm(self, id, *args, **kwargs):
        return self.request.user.is_superuser


@rest_view("/admin/authors")
class AdminAuthorRestView(AuthorRestViewMixin, RestView):
    def perm(self):
        return self.request.user.is_superuser


class BookRestViewMixin():
    model = Book
    fields = [
        'name', 'author', 'pub_date', VF('created_at', readonly=True),
        'type', 'cover', 'pages', 'rating', 'author__last_name',
        'read_time']
    filters = [
        FilterGroup('custom', _("Custom"), [
            ModelFilter(Book, 'author'),
            ModelFilter(Book, 'type'),
            ModelFilter(Book, 'pages'),
            ModelFilter(Book, 'rating'),
            ModelFilter(Book, 'pub_date'),
            Filter(
                'pub_date_gte', _("Publication Date greater than"),
                FieldType.DateField, pub_date_gte_filter),
            Filter(
                'author_pk', _("Author PK"),
                FieldType.ModelObject, author_pk_filter,
                related_model=Author),
        ]),
    ]


@rest_view("/books")
class BookRestView(BookRestViewMixin, MediaRestViewMixin, RestView):
    default_public = True

    def get_queryset(self, *args, **kwargs):
        return super().get_queryset(*args, **kwargs).annotate(
            pages_null=Coalesce('pages', Value(0)))

    def get_order_mapping(self, *order):
        return dict(pages='pages_null')


@rest_view("/perms/books")
class PermsBookRestView(
        BookRestViewMixin, MediaPermsRestViewMixin, PermsRestView):
    pass


def author_json_repr(obj):
    """JSON representation.
    ---
    hello:
        type: string
    last_name:
        type: string
    """
    return obj.json_repr(
        hello="World",
        last_name=obj.last_name)


@rest_view("/books-custom-author")
class BookCustomAuthorRestView(BookRestView):
    fields = [
        VF('name', alias='book_name'),
        VF('author', json_repr=author_json_repr),
        VF('author__last_name', alias='author_last_name'),
        'pub_date']


@rest_view("/book-types")
class BookTypeRestView(RestView):
    model = BookType
    fields = ['name', 'slug']
    default_public = True


@rest_view("/test-i18n")
class Testi18nView(BaseRestView):
    default_public = True

    @rest_api("", method="get")
    def get(self, *args, **kwargs):
        return JsonResponse({'Monday': _("Monday")})


@rest_view("/error")
class TestErrorView(BaseRestView):
    default_public = True

    @rest_api("/500", method="get")
    def raise_500(self):
        raise ImproperlyConfigured("Test exception")


@rest_view("/timezone")
class TestTimezoneView(BaseRestView):
    default_public = True

    @rest_api("", method="get")
    def get(self):
        return JsonResponse({'tz': self.request.TIMEZONE})


@rest_view("/illegal-priority")
class IllegalPriorityRestView(BaseRestView):
    default_public = True

    @rest_api("/test", method="get", priority=1)
    def get(self, *args, **kwargs):
        return JsonResponse(dict(value='test'))

    @rest_api("/test", method="put", priority=2)
    def put(self, *args, **kwargs):
        return JsonResponse(dict(value='test'))


class FakeViewMixin:
    """This mixin allow to create a test view that is not processsed
    by makeapidoc."""

    @classmethod
    def as_urlpatterns(cls):
        return cls.get_urls(as_pattern=True)



from django.test import TestCase

from pfx.pfxcore.models import PFXUser
from pfx.pfxcore.shortcuts import permissions
from pfx.pfxcore.test import APIClient, TestAssertMixin


class TestApi(TestAssertMixin, TestCase):

    def setUp(self):
        self.client = APIClient(default_locale='en')

    @classmethod
    def setUpTestData(cls):
        cls.user = PFXUser.objects.create_user(
            username='user',
            password='test',
            first_name='Test',
            last_name='User',
        )

    def test_get_user(self):
        self.client.login(username='user')

        response = self.client.get(f'/api/users/{self.user.pk}')
        self.assertRC(response, 403)

        self.user.user_permissions.add(*permissions('pfxcore.view_pfxuser'))

        response = self.client.get(f'/api/users/{self.user.pk}')
        self.assertRC(response, 200)
        self.assertJE(response, 'username', 'user')
